# Copyright (c) 2017-2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

from ansible.module_utils.common.text.converters import to_native

from ansible_collections.community.dns.plugins.module_utils._record import DNSRecord
from ansible_collections.community.dns.plugins.module_utils._wsdl import (
    Composer,
    WSDLError,
    WSDLNetworkError,
)
from ansible_collections.community.dns.plugins.module_utils._zone import (
    DNSZone,
    DNSZoneWithRecords,
)
from ansible_collections.community.dns.plugins.module_utils._zone_record_api import (
    NOT_PROVIDED,
    DNSAPIAuthenticationError,
    DNSAPIError,
    ZoneRecordAPI,
    filter_records,
)

if t.TYPE_CHECKING:
    from .._http import HTTPHelper  # pragma: no cover
    from .._record import IDNSRecord  # pragma: no cover
    from .._zone_record_api import NotProvidedType  # pragma: no cover

    _T = t.TypeVar("_T")  # pragma: no cover


def _create_record_from_encoding(
    source: dict[str, t.Any], record_type: str | None = None
) -> DNSRecord[int]:
    source = dict(source)
    record_type = source.pop("type", record_type)
    if record_type is None:
        raise DNSAPIError("Record from API has no type")

    priority = source.pop("priority")
    target: str = source.pop("target")
    if record_type in ("PTR", "MX"):
        target = f"{priority} {target}"

    result = DNSRecord(
        record_id=source.pop("id"), record_type=record_type, target=target
    )
    result.prefix = source.pop("prefix", None)
    ttl = source.pop("ttl")
    result.ttl = int(ttl) if ttl is not None else None
    source.pop("zone", None)
    result.extra["comment"] = source.pop("comment") or ""
    result.extra.update(source)
    return result


def _create_zone_from_encoding(
    source: dict[str, t.Any],
    prefix: str | None | NotProvidedType = NOT_PROVIDED,
    record_type: str | NotProvidedType = NOT_PROVIDED,
) -> DNSZoneWithRecords[int, int]:
    zone = DNSZone(zone_id=source["id"], name=source["name"])
    zone.info = {
        "email": source.get("email"),
        "ttl": source["ttl"],
    }
    return DNSZoneWithRecords(
        zone,
        filter_records(
            [_create_record_from_encoding(record) for record in source["records"]],
            prefix=prefix,
            record_type=record_type,
        ),
    )


def _encode_record(
    record: IDNSRecord[int | None], include_id: bool = False
) -> dict[str, t.Any]:
    result: dict[str, t.Any] = {
        "type": record.type,
        "prefix": record.prefix,
        "target": record.target,
        "ttl": record.ttl,
    }
    if record.type in ("PTR", "MX"):
        try:
            priority, target = record.target.split(" ", 1)
            result["priority"] = int(priority)
            result["target"] = target
        except Exception as e:
            raise DNSAPIError(
                f'Cannot split {record.type} record "{record.target}" into integer priority and target: {e}'
            ) from e
    else:
        result["priority"] = None
    if include_id:
        result["id"] = record.id
    return result


class HostTechWSDLAPI(ZoneRecordAPI[int, int]):
    def __init__(
        self,
        http_helper: HTTPHelper,
        username: str,
        password: str,
        api: str = "https://ns1.hosttech.eu/public/api",
        debug: bool = False,
    ) -> None:
        """
        Create a new HostTech API instance with given username and password.
        """
        self._http_helper = http_helper
        self._api = api
        self._namespaces = {
            "ns1": "https://ns1.hosttech.eu/soap",
        }
        self._username = username
        self._password = password
        self._debug = debug

    def _prepare(self) -> Composer:
        command = Composer(self._http_helper, self._api, self._namespaces)
        command.add_auth(self._username, self._password)
        return command

    def _announce(self, msg: str) -> None:
        if self._debug:
            pass  # pragma: no cover
            # q.q('{0} {1} {2}'.format('=' * 4, msg, '=' * 40))

    def _execute(
        self, command: Composer, result_name: str, acceptable_type: type[_T]
    ) -> _T:
        if self._debug:
            pass  # pragma: no cover
            # q.q('Request: {0}'.format(command))
        try:
            result = command.execute(debug=self._debug)
        except WSDLError as e:
            if e.error_code == "998":
                raise DNSAPIAuthenticationError(
                    f"Error on authentication ({e.error_message})"
                ) from e
            raise
        res = result.get_result(result_name)
        if isinstance(res, acceptable_type):
            if self._debug:
                pass  # pragma: no cover
                # q.q('Extracted result: {0} (type {1})'.format(res, type(res)))
            return res
        if self._debug:
            pass  # pragma: no cover
            # q.q('Result: {0}; extracted type {1}'.format(result, type(res)))
        raise DNSAPIError(
            f"Result has unexpected type {type(res)} (expecting {acceptable_type})!"
        )

    def get_zone_with_records_by_name(
        self,
        name: str,
        prefix: str | None | NotProvidedType = NOT_PROVIDED,
        record_type: str | NotProvidedType = NOT_PROVIDED,
    ) -> DNSZoneWithRecords[int, int] | None:
        """
        Given a zone name, return the zone contents with records if found.

        @param name: The zone name (string)
        @param prefix: The prefix to filter for, if provided. Since None is a valid value,
                       the special constant NOT_PROVIDED indicates that we are not filtering.
        @param record_type: The record type to filter for, if provided
        @return The zone information with records (DNSZoneWithRecords), or None if not found
        """
        self._announce("get zone")
        command = self._prepare()
        command.add_simple_command("getZone", sZoneName=name)
        try:
            return _create_zone_from_encoding(
                self._execute(command, "getZoneResponse", dict),
                prefix=prefix,
                record_type=record_type,
            )
        except WSDLError as exc:
            if exc.error_origin == "server" and exc.error_message == "zone not found":
                return None
            raise DNSAPIError(f"Error while getting zone: {to_native(exc)}") from exc
        except WSDLNetworkError as exc:
            raise DNSAPIError(
                f"Network error while getting zone: {to_native(exc)}"
            ) from exc

    def get_zone_with_records_by_id(
        self,
        zone_id: int,
        prefix: str | None | NotProvidedType = NOT_PROVIDED,
        record_type: str | NotProvidedType = NOT_PROVIDED,
    ) -> DNSZoneWithRecords[int, int] | None:
        """
        Given a zone ID, return the zone contents with records if found.

        @param zone_id: The zone ID
        @param prefix: The prefix to filter for, if provided. Since None is a valid value,
                       the special constant NOT_PROVIDED indicates that we are not filtering.
        @param record_type: The record type to filter for, if provided
        @return The zone information with records (DNSZoneWithRecords), or None if not found
        """
        return self.get_zone_with_records_by_name(
            str(zone_id), prefix=prefix, record_type=record_type
        )

    def get_zone_records(
        self,
        zone_id: int,
        prefix: str | None | NotProvidedType = NOT_PROVIDED,
        record_type: str | NotProvidedType = NOT_PROVIDED,
    ) -> list[DNSRecord[int]] | None:
        """
        Given a zone ID, return a list of records, optionally filtered by the provided criteria.

        @param zone_id: The zone ID
        @param prefix: The prefix to filter for, if provided. Since None is a valid value,
                       the special constant NOT_PROVIDED indicates that we are not filtering.
        @param record_type: The record type to filter for, if provided
        @return A list of DNSrecord objects, or None if zone was not found
        """
        result = self.get_zone_with_records_by_id(
            zone_id, prefix=prefix, record_type=record_type
        )
        return result.records if result is not None else None

    def get_zone_by_name(self, name: str) -> DNSZone[int] | None:
        """
        Given a zone name, return the zone contents if found.

        @param name: The zone name (string)
        @return The zone information (DNSZone), or None if not found
        """
        zone = self.get_zone_with_records_by_name(name)
        return zone.zone if zone else None

    def get_zone_by_id(self, zone_id: int) -> DNSZone[int] | None:
        """
        Given a zone ID, return the zone contents if found.

        @param zone_id: The zone ID
        @return The zone information (DNSZone), or None if not found
        """
        zone = self.get_zone_with_records_by_id(zone_id)
        return zone.zone if zone else None

    def add_record(
        self, zone_id: int, record: IDNSRecord[int | None]
    ) -> DNSRecord[int]:
        """
        Adds a new record to an existing zone.

        @param zone_id: The zone ID
        @param record: The DNS record (DNSRecord)
        @return The created DNS record (DNSRecord)
        """
        self._announce("add record")
        command = self._prepare()
        command.add_simple_command(
            "addRecord",
            search=str(zone_id),
            recorddata=_encode_record(record, include_id=False),
        )
        try:
            return _create_record_from_encoding(
                self._execute(command, "addRecordResponse", dict)
            )
        except WSDLError as exc:
            raise DNSAPIError(f"Error while adding record: {to_native(exc)}") from exc
        except WSDLNetworkError as exc:
            raise DNSAPIError(
                f"Network error while adding record: {to_native(exc)}"
            ) from exc

    def update_record(self, zone_id: int, record: DNSRecord[int]) -> DNSRecord[int]:
        """
        Update a record.

        @param zone_id: The zone ID
        @param record: The DNS record (DNSRecord)
        @return The DNS record (DNSRecord)
        """
        if record.id is None:
            raise DNSAPIError("Need record ID to update record!")
        self._announce("update record")
        command = self._prepare()
        command.add_simple_command(
            "updateRecord",
            recordId=record.id,
            recorddata=_encode_record(record, include_id=False),
        )
        try:
            return _create_record_from_encoding(
                self._execute(command, "updateRecordResponse", dict)
            )
        except WSDLError as exc:
            raise DNSAPIError(f"Error while updating record: {to_native(exc)}") from exc
        except WSDLNetworkError as exc:
            raise DNSAPIError(
                f"Network error while updating record: {to_native(exc)}"
            ) from exc

    def delete_record(self, zone_id: int, record: DNSRecord[int]) -> bool:
        """
        Delete a record.

        @param zone_id: The zone ID
        @param record: The DNS record (DNSRecord)
        @return True in case of success (boolean)
        """
        if record.id is None:
            raise DNSAPIError("Need record ID to delete record!")
        self._announce("delete record")
        command = self._prepare()
        command.add_simple_command("deleteRecord", recordId=record.id)
        try:
            return self._execute(command, "deleteRecordResponse", bool)
        except WSDLError as exc:
            raise DNSAPIError(f"Error while deleting record: {to_native(exc)}") from exc
        except WSDLNetworkError as exc:
            raise DNSAPIError(
                f"Network error while deleting record: {to_native(exc)}"
            ) from exc



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2016, Thomas Stringer <tomstr@microsoft.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_loadbalancer

version_added: "0.1.2"

short_description: Manage Azure load balancers

description:
    - Create, update and delete Azure load balancers.

options:
    resource_group:
        description:
            - Name of a resource group where the load balancer exists or will be created.
        required: true
        type: str
    name:
        description:
            - Name of the load balancer.
        required: true
        type: str
    state:
        description:
            - Assert the state of the load balancer. Use C(present) to create/update a load balancer, or C(absent) to delete one.
        default: present
        type: str
        choices:
            - absent
            - present
    location:
        description:
            - Valid Azure location. Defaults to location of the resource group.
        type: str
    sku:
        description:
            - The load balancer SKU.
        type: str
        choices:
            - Basic
            - Standard
    frontend_ip_configurations:
        description:
            - List of frontend IPs to be used.
        type: list
        elements: dict
        suboptions:
            name:
                description:
                    - Name of the frontend ip configuration.
                type: str
                required: True
            public_ip_address:
                description:
                    - Name of an existing public IP address object in the current resource group to associate with the security group.
                type: str
            private_ip_address:
                description:
                    - The reference of the Public IP resource.
                type: str
            private_ip_allocation_method:
                description:
                    - The Private IP allocation method.
                type: str
                choices:
                    - Static
                    - Dynamic
            subnet:
                description:
                    - The reference of the subnet resource.
                    - Should be an existing subnet's resource id.
                type: str
            zones:
                description:
                    - list of availability zones denoting the IP allocated for the resource needs to come from.
                    - This must be specified I(sku=Standard) and I(subnet) when setting zones.
                type: list
                elements: str
    backend_address_pools:
        description:
            - List of backend address pools.
        type: list
        elements: dict
        suboptions:
            name:
                description:
                    - Name of the backend address pool.
                required: True
                type: str
    probes:
        description:
            - List of probe definitions used to check endpoint health.
        type: list
        elements: dict
        suboptions:
            name:
                description:
                    - Name of the probe.
                type: str
                required: True
            port:
                description:
                    - Probe port for communicating the probe. Possible values range from 1 to 65535, inclusive.
                type: int
                required: True
            protocol:
                description:
                    - The protocol of the end point to be probed.
                    - If C(Tcp) is specified, a received ACK is required for the probe to be successful.
                    - If C(Http) or C(Https) is specified, a 200 OK response from the specified URL is required for the probe to be successful.
                type: str
                choices:
                    - Tcp
                    - Http
                    - Https
            interval:
                description:
                    - The interval, in seconds, for how frequently to probe the endpoint for health status.
                    - Slightly less than half the allocated timeout period, which allows two full probes before taking the instance out of rotation.
                    - The default value is C(15), the minimum value is C(5).
                type: int
                default: 15
            fail_count:
                description:
                    - The number of probes where if no response, will result in stopping further traffic from being delivered to the endpoint.
                    - This values allows endpoints to be taken out of rotation faster or slower than the typical times used in Azure.
                default: 3
                type: int
                aliases:
                    - number_of_probes
            request_path:
                description:
                    - The URI used for requesting health status from the VM.
                    - Path is required if I(protocol=Http) or I(protocol=Https). Otherwise, it is not allowed.
                type: str
    inbound_nat_pools:
        description:
            - Defines an external port range for inbound NAT to a single backend port on NICs associated with a load balancer.
            - Inbound NAT rules are created automatically for each NIC associated with the Load Balancer using an external port from this range.
            - Defining an Inbound NAT pool on your Load Balancer is mutually exclusive with defining inbound Nat rules.
            - Inbound NAT pools are referenced from virtual machine scale sets.
            - NICs that are associated with individual virtual machines cannot reference an inbound NAT pool.
            - They have to reference individual inbound NAT rules.
        type: list
        elements: dict
        suboptions:
            name:
                description:
                    - Name of the inbound NAT pool.
                type: str
                required: True
            frontend_ip_configuration_name:
                description:
                    - A reference to frontend IP addresses.
                required: True
                type: str
            protocol:
                description:
                    - IP protocol for the NAT pool.
                type: str
                choices:
                    - Tcp
                    - Udp
                    - All
            frontend_port_range_start:
                description:
                    - The first port in the range of external ports that will be used to provide inbound NAT to NICs associated with the load balancer.
                    - Acceptable values range between 1 and 65534.
                type: int
                required: True
            frontend_port_range_end:
                description:
                    - The last port in the range of external ports that will be used to provide inbound NAT to NICs associated with the load balancer.
                    - Acceptable values range between 1 and 65535.
                type: int
                required: True
            backend_port:
                description:
                    - The port used for internal connections on the endpoint.
                    - Acceptable values are between 1 and 65535.
                type: int
                required: true
    load_balancing_rules:
        description:
            - Object collection representing the load balancing rules Gets the provisioning.
        type: list
        elements: dict
        suboptions:
            name:
                description:
                    - Name of the load balancing rule.
                type: str
                required: True
            frontend_ip_configuration:
                description:
                    - A reference to frontend IP addresses.
                type: str
                required: True
            backend_address_pool:
                description:
                    - A reference to a pool of DIPs. Inbound traffic is randomly load balanced across IPs in the backend IPs.
                required: True
                type: str
            probe:
                description:
                    - The name of the load balancer probe this rule should use for health checks.
                required: True
                type: str
            protocol:
                description:
                    - IP protocol for the load balancing rule.
                type: str
                choices:
                    - Tcp
                    - Udp
                    - All
            load_distribution:
                description:
                    - The session persistence policy for this rule; C(Default) is no persistence.
                type: str
                choices:
                    - Default
                    - SourceIP
                    - SourceIPProtocol
                default: Default
            frontend_port:
                description:
                    - The port for the external endpoint.
                    - Frontend port numbers must be unique across all rules within the load balancer.
                    - Acceptable values are between 0 and 65534.
                    - Note that value 0 enables "Any Port".
                type: int
                required: true
            backend_port:
                description:
                    - The port used for internal connections on the endpoint.
                    - Acceptable values are between 0 and 65535.
                    - Note that value 0 enables "Any Port".
                type: int
            idle_timeout:
                description:
                    - The timeout for the TCP idle connection.
                    - The value can be set between 4 and 30 minutes.
                    - The default value is C(4) minutes.
                    - This element is only used when the protocol is set to TCP.
                type: int
                default: 4
            enable_floating_ip:
                description:
                    - Configures a virtual machine's endpoint for the floating IP capability required to configure a SQL AlwaysOn Availability Group.
                type: bool
            enable_tcp_reset:
                description:
                    - Receive bidirectional TCP Reset on TCP flow idle timeout or unexpected connection termination.
                    - This element is only used when the protocol is set to TCP.
                type: bool
            disable_outbound_snat:
                description:
                    - Configure outbound source network address translation (SNAT).
                    - The default behavior when omitted is equivalent to I(disable_outbound_snat=True).
                    - True is equivalent to "(Recommended) Use outbound rules to provide backend pool members access to the internet" in portal.
                    - False is equivalent to "Use default outbound access" in portal.
                type: bool
                default: False
    inbound_nat_rules:
        description:
            - Collection of inbound NAT Rules used by a load balancer.
            - Defining inbound NAT rules on your load balancer is mutually exclusive with defining an inbound NAT pool.
            - Inbound NAT pools are referenced from virtual machine scale sets.
            - NICs that are associated with individual virtual machines cannot reference an Inbound NAT pool.
            - They have to reference individual inbound NAT rules.
        type: list
        elements: dict
        suboptions:
            name:
                description:
                    - name of the inbound nat rule.
                type: str
                required: True
            frontend_ip_configuration:
                description:
                    - A reference to frontend IP addresses.
                type: str
                required: True
            protocol:
                description:
                    - IP protocol for the inbound nat rule.
                type: str
                choices:
                    - Tcp
                    - Udp
                    - All
            frontend_port:
                description:
                    - The port for the external endpoint.
                    - Frontend port numbers must be unique across all rules within the load balancer.
                    - Acceptable values are between 0 and 65534.
                    - Note that value 0 enables "Any Port".
                type: int
                required: True
            backend_port:
                description:
                    - The port used for internal connections on the endpoint.
                    - Acceptable values are between 0 and 65535.
                    - Note that value 0 enables "Any Port".
                type: int
                required: true
            idle_timeout:
                description:
                    - The timeout for the TCP idle connection.
                    - The value can be set between 4 and 30 minutes.
                    - The default value is C(4) minutes.
                    - This element is only used when I(protocol=Tcp).
                type: int
            enable_floating_ip:
                description:
                    - Configures a virtual machine's endpoint for the floating IP capability required to configure a SQL AlwaysOn Availability Group.
                    - This setting is required when using the SQL AlwaysOn Availability Groups in SQL server.
                    - This setting can't be changed after you create the endpoint.
                type: bool
            enable_tcp_reset:
                description:
                    - Receive bidirectional TCP Reset on TCP flow idle timeout or unexpected connection termination.
                    - This element is only used when I(protocol=Tcp).
                type: bool
    outbound_rules:
        description:
            - The outbound rules.
        type: list
        elements: dict
        version_added: '3.5.0'
        suboptions:
            name:
                description:
                    - The name of the resource that is unique within the set of outbound rules used by the load balancer.
                    - This name can be used to access the resource.
                type: str
                required: True
            allocated_outbound_ports:
                description:
                    - The number of outbound ports to be used for NAT.
                type: int
            frontend_ip_configurations:
                description:
                    - The Frontend IP addresses of the load balancer.
                type: list
                elements: str
            backend_address_pool:
                description:
                    - A reference to a pool of DIPs.
                    - Outbound traffic is randomly load balanced across IPs in the backend IPs.
                type: str
            protocol:
                description:
                    - The protocol for the outbound rule in load balancer.
                type: str
                choices:
                    - Tcp
                    - Udp
                    - All
            enable_tcp_reset:
                description:
                    - Receive bidirectional TCP Reset on TCP flow idle timeout or unexpected connection termination.
                    - This element is only used when the protocol is set to C(Tcp).
                type: bool
            idle_timeout_in_minutes:
                description:
                    - The timeout for the TCP idle connection.
                type: int
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Thomas Stringer (@trstringer)
    - Yuwei Zhou (@yuwzho)
'''

EXAMPLES = '''
- name: create load balancer
  azure_rm_loadbalancer:
    resource_group: myResourceGroup
    name: testloadbalancer1
    frontend_ip_configurations:
      - name: frontendipconf0
        public_ip_address: testpip
      - name: frontendipconf1
        public_ip_address: testpip1
    backend_address_pools:
      - name: backendaddrpool0
      - name: backendaddrpool1
    probes:
      - name: prob0
        port: 80
    inbound_nat_pools:
      - name: inboundnatpool0
        frontend_ip_configuration_name: frontendipconf0
        protocol: Tcp
        frontend_port_range_start: 80
        frontend_port_range_end: 81
        backend_port: 8080
    load_balancing_rules:
      - name: lbrbalancingrule0
        frontend_ip_configuration: frontendipconf0
        backend_address_pool: backendaddrpool0
        frontend_port: 80
        backend_port: 80
        probe: prob0
    inbound_nat_rules:
      - name: inboundnatrule0
        backend_port: 8080
        protocol: Tcp
        frontend_port: 8080
        frontend_ip_configuration: frontendipconf0
    outbound_rules:
      - name: outrule1
        allocated_outbound_ports: 800
        frontend_ip_configurations:
          - frontendipconf1
        backend_address_pool: backendaddrpool1
        protocol: Tcp
        enable_tcp_reset: true
        idle_timeout_in_minutes: 4
'''

RETURN = '''
state:
    description:
        - Current state of the load balancer.
    returned: always
    type: dict
changed:
    description:
        - Whether or not the resource has changed.
    returned: always
    type: bool
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import format_resource_id
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt
from ansible.module_utils._text import to_native
try:
    from azure.core.exceptions import ResourceNotFoundError
    from azure.mgmt.core.tools import parse_resource_id
except ImportError:
    # This is handled in azure_rm_common
    pass


frontend_ip_configuration_spec = dict(
    name=dict(
        type='str',
        required=True
    ),
    public_ip_address=dict(
        type='str'
    ),
    private_ip_address=dict(
        type='str'
    ),
    private_ip_allocation_method=dict(
        type='str',
        choices=['Static', 'Dynamic']
    ),
    subnet=dict(
        type='str'
    ),
    zones=dict(
        type='list',
        elements='str'
    )
)


backend_address_pool_spec = dict(
    name=dict(
        type='str',
        required=True
    )
)


probes_spec = dict(
    name=dict(
        type='str',
        required=True
    ),
    port=dict(
        type='int',
        required=True
    ),
    protocol=dict(
        type='str',
        choices=['Tcp', 'Http', 'Https']
    ),
    interval=dict(
        type='int',
        default=15
    ),
    fail_count=dict(
        type='int',
        default=3,
        aliases=['number_of_probes']
    ),
    request_path=dict(
        type='str'
    )
)


inbound_nat_pool_spec = dict(
    name=dict(
        type='str',
        required=True
    ),
    frontend_ip_configuration_name=dict(
        type='str',
        required=True
    ),
    protocol=dict(
        type='str',
        choices=['Tcp', 'Udp', 'All']
    ),
    frontend_port_range_start=dict(
        type='int',
        required=True
    ),
    frontend_port_range_end=dict(
        type='int',
        required=True
    ),
    backend_port=dict(
        type='int',
        required=True
    )
)


inbound_nat_rule_spec = dict(
    name=dict(
        type='str',
        required=True
    ),
    frontend_ip_configuration=dict(
        type='str',
        required=True
    ),
    protocol=dict(
        type='str',
        choices=['Tcp', 'Udp', 'All']
    ),
    frontend_port=dict(
        type='int',
        required=True
    ),
    idle_timeout=dict(
        type='int'
    ),
    backend_port=dict(
        type='int',
        required=True
    ),
    enable_floating_ip=dict(
        type='bool'
    ),
    enable_tcp_reset=dict(
        type='bool'
    )
)


load_balancing_rule_spec = dict(
    name=dict(
        type='str',
        required=True
    ),
    frontend_ip_configuration=dict(
        type='str',
        required=True
    ),
    backend_address_pool=dict(
        type='str',
        required=True
    ),
    probe=dict(
        type='str',
        required=True
    ),
    protocol=dict(
        type='str',
        choices=['Tcp', 'Udp', 'All']
    ),
    load_distribution=dict(
        type='str',
        choices=['Default', 'SourceIP', 'SourceIPProtocol'],
        default='Default'
    ),
    frontend_port=dict(
        type='int',
        required=True
    ),
    backend_port=dict(
        type='int'
    ),
    idle_timeout=dict(
        type='int',
        default=4
    ),
    enable_floating_ip=dict(
        type='bool'
    ),
    disable_outbound_snat=dict(
        type='bool',
        default=False
    ),
    enable_tcp_reset=dict(
        type='bool'
    )
)


outbound_rule_spec = dict(
    name=dict(type='str', required=True),
    allocated_outbound_ports=dict(type='int'),
    frontend_ip_configurations=dict(type='list', elements='str'),
    backend_address_pool=dict(type='str'),
    protocol=dict(type='str', choices=['Tcp', 'Udp', 'All']),
    enable_tcp_reset=dict(type='bool'),
    idle_timeout_in_minutes=dict(type='int')

)


class AzureRMLoadBalancer(AzureRMModuleBaseExt):
    """Configuration class for an Azure RM load balancer resource"""

    def __init__(self):
        self.module_args = dict(
            resource_group=dict(
                type='str',
                required=True
            ),
            name=dict(
                type='str',
                required=True
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            ),
            location=dict(
                type='str'
            ),
            sku=dict(
                type='str',
                choices=['Basic', 'Standard']
            ),
            frontend_ip_configurations=dict(
                type='list',
                elements='dict',
                options=frontend_ip_configuration_spec
            ),
            backend_address_pools=dict(
                type='list',
                elements='dict',
                options=backend_address_pool_spec
            ),
            probes=dict(
                type='list',
                elements='dict',
                options=probes_spec
            ),
            inbound_nat_rules=dict(
                type='list',
                elements='dict',
                options=inbound_nat_rule_spec
            ),
            inbound_nat_pools=dict(
                type='list',
                elements='dict',
                options=inbound_nat_pool_spec
            ),
            load_balancing_rules=dict(
                type='list',
                elements='dict',
                options=load_balancing_rule_spec
            ),
            outbound_rules=dict(
                type='list',
                elements='dict',
                options=outbound_rule_spec
            )
        )

        self.resource_group = None
        self.name = None
        self.location = None
        self.sku = None
        self.frontend_ip_configurations = None
        self.backend_address_pools = None
        self.probes = None
        self.inbound_nat_rules = None
        self.inbound_nat_pools = None
        self.outbound_rules = None
        self.load_balancing_rules = None
        self.state = None
        self.tags = None

        self.results = dict(changed=False, state=dict())

        super(AzureRMLoadBalancer, self).__init__(
            derived_arg_spec=self.module_args,
            supports_check_mode=True
        )

    def exec_module(self, **kwargs):
        """Main module execution method"""
        for key in list(self.module_args.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        changed = False

        resource_group = self.get_resource_group(self.resource_group)
        if not self.location:
            self.location = resource_group.location

        load_balancer = self.get_load_balancer()

        if self.state == 'present':
            # create new load balancer structure early, so it can be easily compared
            if not load_balancer:
                changed = True
            else:
                response = self.format_item(load_balancer)
                update1, self.frontend_ip_configurations = self.update_item(self.frontend_ip_configurations, response.get('frontend_ip_configurations'))
                update2, self.backend_address_pools = self.update_item(self.backend_address_pools, response.get('backend_address_pools'))
                update3, self.probes = self.update_item(self.probes, response.get('probes'))
                update4, self.inbound_nat_rules = self.update_item(self.inbound_nat_rules, response.get('inbound_nat_rules'))
                update5, self.inbound_nat_pools = self.update_item(self.inbound_nat_pools, response.get('inbound_nat_pools'))
                update6, self.outbound_rules = self.update_item(self.outbound_rules, response.get('outbound_rules'))
                update7, self.load_balancing_rules = self.update_item(self.load_balancing_rules, response.get('load_balancing_rules'))

                changed = update1 or update2 or update3 or update4 or update5 or update6 or update7

            frontend_ip_configurations_param = [self.network_models.FrontendIPConfiguration(
                name=item.get('name'),
                public_ip_address=self.get_public_ip_address_instance(item.get('public_ip_address')) if item.get('public_ip_address') else None,
                private_ip_address=item.get('private_ip_address'),
                private_ip_allocation_method=item.get('private_ip_allocation_method'),
                zones=item.get('zones'),
                subnet=self.network_models.Subnet(
                    id=item.get('subnet'),
                    private_endpoint_network_policies=None,
                    private_link_service_network_policies=None
                ) if item.get('subnet') else None
            ) for item in self.frontend_ip_configurations] if self.frontend_ip_configurations else None

            backend_address_pools_param = [self.network_models.BackendAddressPool(
                name=item.get('name')
            ) for item in self.backend_address_pools] if self.backend_address_pools else None

            probes_param = [self.network_models.Probe(
                name=item.get('name'),
                port=item.get('port'),
                protocol=item.get('protocol'),
                interval_in_seconds=item.get('interval'),
                request_path=item.get('request_path'),
                number_of_probes=item.get('fail_count')
            ) for item in self.probes] if self.probes else None

            inbound_nat_pools_param = [self.network_models.InboundNatPool(
                name=item.get('name'),
                frontend_ip_configuration=self.network_models.SubResource(
                    id=frontend_ip_configuration_id(
                        self.subscription_id,
                        self.resource_group,
                        self.name,
                        item.get('frontend_ip_configuration_name'))),
                protocol=item.get('protocol'),
                frontend_port_range_start=item.get('frontend_port_range_start'),
                frontend_port_range_end=item.get('frontend_port_range_end'),
                backend_port=item.get('backend_port')
            ) for item in self.inbound_nat_pools] if self.inbound_nat_pools else None

            load_balancing_rules_param = [self.network_models.LoadBalancingRule(
                name=item.get('name'),
                frontend_ip_configuration=self.network_models.SubResource(
                    id=frontend_ip_configuration_id(
                        self.subscription_id,
                        self.resource_group,
                        self.name,
                        item.get('frontend_ip_configuration')
                    )
                ),
                backend_address_pool=self.network_models.SubResource(
                    id=backend_address_pool_id(
                        self.subscription_id,
                        self.resource_group,
                        self.name,
                        item.get('backend_address_pool')
                    )
                ),
                probe=self.network_models.SubResource(
                    id=probe_id(
                        self.subscription_id,
                        self.resource_group,
                        self.name,
                        item.get('probe')
                    )
                ),
                protocol=item.get('protocol'),
                load_distribution=item.get('load_distribution'),
                frontend_port=item.get('frontend_port'),
                backend_port=item.get('backend_port'),
                idle_timeout_in_minutes=item.get('idle_timeout'),
                enable_floating_ip=item.get('enable_floating_ip'),
                enable_tcp_reset=item.get('enable_tcp_reset'),
                disable_outbound_snat=item.get('disable_outbound_snat'),
            ) for item in self.load_balancing_rules] if self.load_balancing_rules else None

            inbound_nat_rules_param = [self.network_models.InboundNatRule(
                name=item.get('name'),
                frontend_ip_configuration=self.network_models.SubResource(
                    id=frontend_ip_configuration_id(
                        self.subscription_id,
                        self.resource_group,
                        self.name,
                        item.get('frontend_ip_configuration')
                    )
                ) if item.get('frontend_ip_configuration') else None,
                protocol=item.get('protocol'),
                frontend_port=item.get('frontend_port'),
                backend_port=item.get('backend_port'),
                idle_timeout_in_minutes=item.get('idle_timeout'),
                enable_tcp_reset=item.get('enable_tcp_reset'),
                enable_floating_ip=item.get('enable_floating_ip')
            ) for item in self.inbound_nat_rules] if self.inbound_nat_rules else None

            outbound_rules_param = [self.network_models.OutboundRule(
                name=item.get('name'),
                frontend_ip_configurations=[self.network_models.SubResource(
                    id=frontend_ip_configuration_id(
                        self.subscription_id,
                        self.resource_group,
                        self.name,
                        value
                    )
                )for value in item['frontend_ip_configurations']] if item.get('frontend_ip_configurations') else None,
                backend_address_pool=self.network_models.SubResource(
                    id=backend_address_pool_id(
                        self.subscription_id,
                        self.resource_group,
                        self.name,
                        item.get('backend_address_pool')
                    )
                ),
                allocated_outbound_ports=item.get('allocated_outbound_ports'),
                protocol=item.get('protocol'),
                enable_tcp_reset=item.get('enable_tcp_reset'),
                idle_timeout_in_minutes=item.get('idle_timeout_in_minutes')
            ) for item in self.outbound_rules] if self.outbound_rules else None

            # construct the new instance, if the parameter is none, keep remote one
            self.new_load_balancer = self.network_models.LoadBalancer(
                sku=self.network_models.LoadBalancerSku(name=self.sku) if self.sku else None,
                location=self.location,
                tags=self.tags,
                frontend_ip_configurations=frontend_ip_configurations_param,
                backend_address_pools=backend_address_pools_param,
                probes=probes_param,
                inbound_nat_pools=inbound_nat_pools_param,
                load_balancing_rules=load_balancing_rules_param,
                inbound_nat_rules=inbound_nat_rules_param,
                outbound_rules=outbound_rules_param,
            )

            self.new_load_balancer = self.assign_protocol(self.new_load_balancer, load_balancer)

            if load_balancer:
                self.new_load_balancer = self.object_assign(self.new_load_balancer, load_balancer)
        elif self.state == 'absent' and load_balancer:
            changed = True

        self.results['state'] = load_balancer.as_dict() if load_balancer else {}
        if 'tags' in self.results['state']:
            update_tags, self.results['state']['tags'] = self.update_tags(self.results['state']['tags'])
            if update_tags:
                changed = True
        else:
            if self.tags:
                changed = True
        self.results['changed'] = changed

        if self.check_mode:
            return self.results

        if self.state == 'present' and changed:
            self.results['state'] = self.create_or_update_load_balancer(self.new_load_balancer).as_dict()
        elif self.state == 'absent' and changed:
            self.delete_load_balancer()
            self.results['state'] = None

        return self.results

    def update_item(self, new, old):
        changed = False
        if new is not None:
            if old is not None:
                if not self.default_compare({}, new, old, '', dict(compare=[])):
                    changed = True

                keys = [item['name'] for item in new]
                for item in old:
                    if item['name'] not in keys:
                        new.append(item)
            else:
                changed = True
        else:
            new = old
        return changed, new

    def format_item(self, item):
        results = dict()
        if item.frontend_ip_configurations is not None:
            results['frontend_ip_configurations'] = []
            for value in item.frontend_ip_configurations:
                new_item = dict(name=value.name,
                                public_ip_address=None,
                                private_ip_address=value.private_ip_address,
                                private_ip_allocation_method=value.private_ip_allocation_method,
                                subnet=None,
                                zones=value.zones)
                if value.public_ip_address is not None:
                    new_item['public_ip_address'] = parse_resource_id(value.public_ip_address.id)['name']
                if value.subnet is not None:
                    new_item['subnet'] = value.subnet.id
                results['frontend_ip_configurations'].append(new_item)
        else:
            results['frontend_ip_configurations'] = None
        if item.backend_address_pools is not None:
            results['backend_address_pools'] = []
            for value in item.backend_address_pools:
                results['backend_address_pools'].append(dict(name=value.name))
        else:
            results['backend_address_pools'] = None
        if item.probes is not None:
            results['probes'] = []
            for value in item.probes:
                results['probes'].append(dict(name=value.name,
                                              port=value.port,
                                              protocol=value.protocol,
                                              interval=value.interval_in_seconds,
                                              fail_count=value.number_of_probes,
                                              request_path=value.request_path))
        else:
            results['probes'] = None
        if item.inbound_nat_pools is not None:
            results['inbound_nat_pools'] = []
            for value in item.inbound_nat_pools:
                new_item = dict(name=value.name,
                                frontend_ip_configuration_name=None,
                                protocol=value.protocol,
                                frontend_port_range_start=value.frontend_port_range_start,
                                frontend_port_range_end=value.frontend_port_range_end,
                                backend_port=value.backend_port)
                if value.frontend_ip_configuration is not None:
                    new_item['frontend_ip_configuration_name'] = parse_resource_id(value.frontend_ip_configuration.id)['resource_name']
                results['inbound_nat_pools'].append(new_item)
        else:
            results['inbound_nat_pools'] = None
        if item.inbound_nat_rules is not None:
            results['inbound_nat_rules'] = []
            for value in item.inbound_nat_rules:
                new_item = dict(name=value.name,
                                frontend_ip_configuration=None,
                                protocol=value.protocol,
                                frontend_port=value.frontend_port,
                                idle_timeout=value.idle_timeout_in_minutes,
                                backend_port=value.backend_port,
                                enable_floating_ip=value.enable_floating_ip,
                                enable_tcp_reset=value.enable_tcp_reset)
                if value.frontend_ip_configuration is not None:
                    new_item['frontend_ip_configuration'] = parse_resource_id(value.frontend_ip_configuration.id)['resource_name']
                results['inbound_nat_rules'].append(new_item)
        else:
            results['inbound_nat_rules'] = None
        if item.load_balancing_rules is not None:
            results['load_balancing_rules'] = []
            for value in item.load_balancing_rules:
                new_item = dict(name=value.name,
                                frontend_ip_configuration=None,
                                backend_address_pool=None,
                                probe=None,
                                protocol=value.protocol,
                                load_distribution=value.load_distribution,
                                frontend_port=value.frontend_port,
                                backend_port=value.backend_port,
                                idle_timeout=value.idle_timeout_in_minutes,
                                enable_floating_ip=value.enable_floating_ip,
                                disable_outbound_snat=value.disable_outbound_snat,
                                enable_tcp_reset=value.enable_tcp_reset)
                if value.frontend_ip_configuration is not None:
                    new_item['frontend_ip_configuration'] = parse_resource_id(value.frontend_ip_configuration.id)['resource_name']
                if value.backend_address_pool is not None:
                    new_item['backend_address_pool'] = parse_resource_id(value.backend_address_pool.id)['resource_name']
                if value.probe is not None:
                    new_item['probe'] = parse_resource_id(value.probe.id)['resource_name']
                results['load_balancing_rules'].append(new_item)
        else:
            results['load_balancing_rules'] = None
        if item.outbound_rules is not None:
            results['outbound_rules'] = []
            for value in item.outbound_rules:
                new_item = dict(name=value.name,
                                allocated_outbound_ports=value.allocated_outbound_ports,
                                frontend_ip_configurations=[],
                                backend_address_pool=None,
                                protocol=value.protocol,
                                enable_tcp_reset=value.enable_tcp_reset,
                                idle_timeout_in_minutes=value.idle_timeout_in_minutes)
                if value.frontend_ip_configurations is not None:
                    for key in value.frontend_ip_configurations:
                        new_item['frontend_ip_configurations'].append(parse_resource_id(key.id)['resource_name'])
                else:
                    new_item['frontend_ip_configurations'] = None
                if value.backend_address_pool is not None:
                    new_item['backend_address_pool'] = parse_resource_id(value.backend_address_pool.id)['resource_name']
                results['outbound_rules'].append(new_item)
        else:
            results['outbound_rules'] = None

        return results

    def get_public_ip_address_instance(self, id):
        """Get a reference to the public ip address resource"""
        self.log('Fetching public ip address {0}'.format(id))
        resource_id = format_resource_id(id, self.subscription_id, 'Microsoft.Network', 'publicIPAddresses', self.resource_group)
        return self.network_models.PublicIPAddress(id=resource_id)

    def get_load_balancer(self):
        """Get a load balancer"""
        self.log('Fetching loadbalancer {0}'.format(self.name))
        try:
            return self.network_client.load_balancers.get(self.resource_group, self.name)
        except ResourceNotFoundError:
            return None

    def delete_load_balancer(self):
        """Delete a load balancer"""
        self.log('Deleting loadbalancer {0}'.format(self.name))
        try:
            poller = self.network_client.load_balancers.begin_delete(self.resource_group, self.name)
            return self.get_poller_result(poller)
        except Exception as exc:
            self.fail("Error deleting loadbalancer {0} - {1}".format(self.name, str(exc)))

    def create_or_update_load_balancer(self, param):
        try:
            poller = self.network_client.load_balancers.begin_create_or_update(self.resource_group, self.name, param)
            new_lb = self.get_poller_result(poller)
            return new_lb
        except Exception as exc:
            self.fail("Error creating or updating load balancer {0} - {1}".format(self.name, str(exc)))

    def object_assign(self, patch, origin):
        attribute_map = set(self.network_models.LoadBalancer._attribute_map.keys()) - set(self.network_models.LoadBalancer._validation.keys())
        for key in attribute_map:
            if not getattr(patch, key):
                setattr(patch, key, getattr(origin, key))
        return patch

    def assign_protocol(self, patch, origin):
        attribute_map = ['probes', 'inbound_nat_rules', 'inbound_nat_pools', 'load_balancing_rules']
        for attribute in attribute_map:
            properties = getattr(patch, attribute)
            if not properties:
                continue
            references = getattr(origin, attribute) if origin else []
            for item in properties:
                if item.protocol:
                    continue
                refs = [x for x in references if to_native(x.name) == item.name]
                ref = refs[0] if len(refs) > 0 else None
                item.protocol = ref.protocol if ref else 'Tcp'
        return patch


def frontend_ip_configuration_id(subscription_id, resource_group_name, load_balancer_name, name):
    """Generate the id for a frontend ip configuration"""
    return '/subscriptions/{0}/resourceGroups/{1}/providers/Microsoft.Network/loadBalancers/{2}/frontendIPConfigurations/{3}'.format(
        subscription_id,
        resource_group_name,
        load_balancer_name,
        name
    )


def backend_address_pool_id(subscription_id, resource_group_name, load_balancer_name, name):
    """Generate the id for a backend address pool"""
    return '/subscriptions/{0}/resourceGroups/{1}/providers/Microsoft.Network/loadBalancers/{2}/backendAddressPools/{3}'.format(
        subscription_id,
        resource_group_name,
        load_balancer_name,
        name
    )


def probe_id(subscription_id, resource_group_name, load_balancer_name, name):
    """Generate the id for a probe"""
    return '/subscriptions/{0}/resourceGroups/{1}/providers/Microsoft.Network/loadBalancers/{2}/probes/{3}'.format(
        subscription_id,
        resource_group_name,
        load_balancer_name,
        name
    )


def main():
    """Main execution"""
    AzureRMLoadBalancer()


if __name__ == '__main__':
    main()



# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: x509_crl_info
short_description: Retrieve information from X.509 CRLs in PEM format
version_added: 2.10.0
author:
  - Felix Fontein (@felixfontein)
description:
  - Provided a X.509 crl in PEM format, retrieve information.
  - This is a filter version of the M(community.crypto.x509_crl_info) module.
options:
  _input:
    description:
      - The content of the X.509 CRL in PEM format.
    type: string
    required: true
  list_revoked_certificates:
    description:
      - If set to V(false), the list of revoked certificates is not included in the result.
      - This is useful when retrieving information on large CRL files. Enumerating all revoked certificates can take some
        time, including serializing the result as JSON, sending it to the Ansible controller, and decoding it again.
    type: bool
    default: true
    version_added: 1.7.0
extends_documentation_fragment:
  - community.crypto._name_encoding
seealso:
  - module: community.crypto.x509_crl_info
  - plugin: community.crypto.to_serial
    plugin_type: filter
"""

EXAMPLES = r"""
---
- name: Show the Organization Name of the CRL's subject
  ansible.builtin.debug:
    msg: >-
      {{
        (
          lookup('ansible.builtin.file', '/path/to/cert.pem')
          | community.crypto.x509_crl_info
        ).issuer.organizationName
      }}
"""

RETURN = r"""
_value:
  description:
    - Information on the CRL.
  type: dict
  contains:
    format:
      description:
        - Whether the CRL is in PEM format (V(pem)) or in DER format (V(der)).
      returned: success
      type: str
      sample: pem
      choices:
        - pem
        - der
    issuer:
      description:
        - The CRL's issuer.
        - Note that for repeated values, only the last one will be returned.
        - See O(name_encoding) for how IDNs are handled.
      returned: success
      type: dict
      sample: {"organizationName": "Ansible", "commonName": "ca.example.com"}
    issuer_ordered:
      description: The CRL's issuer as an ordered list of tuples.
      returned: success
      type: list
      elements: list
      sample: [["organizationName", "Ansible"], ["commonName": "ca.example.com"]]
    last_update:
      description: The point in time from which this CRL can be trusted as ASN.1 TIME.
      returned: success
      type: str
      sample: '20190413202428Z'
    next_update:
      description: The point in time from which a new CRL will be issued and the client has to check for it as ASN.1 TIME.
      returned: success
      type: str
      sample: '20190413202428Z'
    digest:
      description: The signature algorithm used to sign the CRL.
      returned: success
      type: str
      sample: sha256WithRSAEncryption
    revoked_certificates:
      description: List of certificates to be revoked.
      returned: success if O(list_revoked_certificates=true)
      type: list
      elements: dict
      contains:
        serial_number:
          description:
            - Serial number of the certificate.
            - This return value is an B(integer). If you need the serial numbers as a colon-separated hex string, such as
              C(11:22:33), you need to convert it to that form with P(community.crypto.to_serial#filter).
          type: int
          sample: 1234
        revocation_date:
          description: The point in time the certificate was revoked as ASN.1 TIME.
          type: str
          sample: '20190413202428Z'
        issuer:
          description:
            - The certificate's issuer.
            - See O(name_encoding) for how IDNs are handled.
          type: list
          elements: str
          sample: ["DNS:ca.example.org"]
        issuer_critical:
          description: Whether the certificate issuer extension is critical.
          type: bool
          sample: false
        reason:
          description:
            - The value for the revocation reason extension.
          type: str
          sample: key_compromise
          choices:
            - unspecified
            - key_compromise
            - ca_compromise
            - affiliation_changed
            - superseded
            - cessation_of_operation
            - certificate_hold
            - privilege_withdrawn
            - aa_compromise
            - remove_from_crl
        reason_critical:
          description: Whether the revocation reason extension is critical.
          type: bool
          sample: false
        invalidity_date:
          description: |-
            The point in time it was known/suspected that the private key was compromised
            or that the certificate otherwise became invalid as ASN.1 TIME.
          type: str
          sample: '20190413202428Z'
        invalidity_date_critical:
          description: Whether the invalidity date extension is critical.
          type: bool
          sample: false
"""

import base64
import binascii
import typing as t
from collections.abc import Callable

from ansible.errors import AnsibleFilterError
from ansible.module_utils.common.text.converters import to_bytes, to_text

from ansible_collections.community.crypto.plugins.module_utils._crypto.basic import (
    OpenSSLObjectError,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.module_backends.crl_info import (
    get_crl_info,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.pem import (
    identify_pem_format,
)
from ansible_collections.community.crypto.plugins.plugin_utils._filter_module import (
    FilterModuleMock,
)


def x509_crl_info_filter(
    data: str | bytes,
    name_encoding: t.Literal["ignore", "idna", "unicode"] = "ignore",
    list_revoked_certificates: bool = True,
) -> dict[str, t.Any]:
    """Extract information from X.509 PEM certificate."""
    if not isinstance(data, (str, bytes)):
        raise AnsibleFilterError(
            f"The community.crypto.x509_crl_info input must be a text type, not {type(data)}"
        )
    if not isinstance(name_encoding, (str, bytes)):
        raise AnsibleFilterError(
            f"The name_encoding option must be of a text type, not {type(name_encoding)}"
        )
    if not isinstance(list_revoked_certificates, bool):
        raise AnsibleFilterError(
            f"The list_revoked_certificates option must be a boolean, not {type(list_revoked_certificates)}"
        )
    name_encoding = t.cast(
        t.Literal["ignore", "idna", "unicode"], to_text(name_encoding)
    )
    if name_encoding not in ("ignore", "idna", "unicode"):
        raise AnsibleFilterError(
            f'The name_encoding option must be one of the values "ignore", "idna", or "unicode", not "{name_encoding}"'
        )

    data_bytes = to_bytes(data)
    if not identify_pem_format(data_bytes):
        try:
            data_bytes = base64.b64decode(to_text(data_bytes))
        except (binascii.Error, TypeError, ValueError, UnicodeEncodeError):
            pass

    module = FilterModuleMock({"name_encoding": name_encoding})
    try:
        return get_crl_info(
            module=module,
            content=data_bytes,
            list_revoked_certificates=list_revoked_certificates,
        )
    except OpenSSLObjectError as exc:
        raise AnsibleFilterError(str(exc)) from exc


class FilterModule:
    """Ansible jinja2 filters"""

    def filters(self) -> dict[str, Callable]:
        return {
            "x509_crl_info": x509_crl_info_filter,
        }



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2016, Ansible, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_certificate_info
short_description: Get information on certificates from a Windows Certificate Store
description:
- Returns information about certificates in a Windows Certificate Store.
version_added: 2.6.0
options:
  thumbprint:
    description:
    - The thumbprint as a hex string of a certificate to find.
    - When specified, filters the I(certificates) return value to a single certificate
    - See the examples for how to format the thumbprint.
    type: str
    required: no
  store_name:
    description:
    - The name of the store to search.
    - See U(https://docs.microsoft.com/en-us/dotnet/api/system.security.cryptography.x509certificates.storename)
      for a list of built-in store names.
    type: str
    default: My
  store_location:
    description:
    - The location of the store to search.
    type: str
    choices: [ CurrentUser, LocalMachine ]
    default: LocalMachine
seealso:
- module: ansible.windows.win_certificate_store
author:
- Micah Hunsberger (@mhunsber)
'''

EXAMPLES = r'''
- name: Obtain information about a particular certificate in the computer's personal store
  ansible.windows.win_certificate_info:
    thumbprint: BD7AF104CF1872BDB518D95C9534EA941665FD27
  register: mycert

# thumbprint can also be lower case
- name: Obtain information about a particular certificate in the computer's personal store
  ansible.windows.win_certificate_info:
    thumbprint: bd7af104cf1872bdb518d95c9534ea941665fd27
  register: mycert

- name: Obtain information about all certificates in the root store
  ansible.windows.win_certificate_info:
    store_name: Root
  register: ca

# Import a pfx and then get information on the certificates
- name: Import pfx certificate that is password protected
  ansible.windows.win_certificate_store:
    path: C:\Temp\cert.pfx
    state: present
    password: VeryStrongPasswordHere!
  become: true
  become_method: runas
  register: mycert

- name: Obtain information on each certificate that was touched
  ansible.windows.win_certificate_info:
    thumbprint: "{{ item }}"
  register: mycert_stats
  loop: "{{ mycert.thumbprints }}"
'''

RETURN = r'''
exists:
  description:
    - Whether any certificates were found in the store.
    - When I(thumbprint) is specified, returns true only if the certificate mathing the thumbprint exists.
  returned: success
  type: bool
  sample: true
certificates:
  description:
    - A list of information about certificates found in the store, sorted by thumbprint.
  returned: success
  type: list
  elements: dict
  contains:
    archived:
      description: Indicates that the certificate is archived.
      type: bool
      sample: false
    dns_names:
      description: Lists the registered dns names for the certificate.
      type: list
      elements: str
      sample: [ '*.m.wikiquote.org', '*.wikipedia.org' ]
    extensions:
      description: The collection of the certificates extensions.
      type: list
      elements: dict
      sample: [
            {
                "critical": false,
                "field": "Subject Key Identifier",
                "value": "88 27 17 09 a9 b6 18 60 8b ec eb ba f6 47 59 c5 52 54 a3 b7"
            },
            {
                "critical": true,
                "field": "Basic Constraints",
                "value": "Subject Type=CA, Path Length Constraint=None"
            },
            {
                "critical": false,
                "field": "Authority Key Identifier",
                "value": "KeyID=2b d0 69 47 94 76 09 fe f4 6b 8d 2e 40 a6 f7 47 4d 7f 08 5e"
            },
            {
                "critical": false,
                "field": "CRL Distribution Points",
                "value": "[1]CRL Distribution Point: Distribution Point Name:Full Name:URL=http://crl.apple.com/root.crl"
            },
            {
                "critical": true,
                "field": "Key Usage",
                "value": "Digital Signature, Certificate Signing, Off-line CRL Signing, CRL Signing (86)"
            },
            {
                "critical": false,
                "field": null,
                "value": "05 00"
            }
        ]
    friendly_name:
      description: The associated alias for the certificate.
      type: str
      sample: Microsoft Root Authority
    has_private_key:
      description: Indicates that the certificate contains a private key.
      type: bool
      sample: false
    intended_purposes:
      description: lists the intended applications for the certificate.
      returned: enhanced key usages extension exists.
      type: list
      sample: [ "Server Authentication" ]
    is_ca:
      description: Indicates that the certificate is a certificate authority (CA) certificate.
      returned: basic constraints extension exists.
      type: bool
      sample: true
    issued_by:
      description: The certificate issuer's common name.
      type: str
      sample: Apple Root CA
    issued_to:
      description: The certificate's common name.
      type: str
      sample: Apple Worldwide Developer Relations Certification Authority
    issuer:
      description: The certificate issuer's distinguished name.
      type: str
      sample: 'CN=Apple Root CA, OU=Apple Certification Authority, O=Apple Inc., C=US'
    key_usages:
      description:
        - Defines how the certificate key can be used.
        - If this value is not defined, the key can be used for any purpose.
      returned: key usages extension exists.
      type: list
      elements: str
      sample: [ "CrlSign", "KeyCertSign", "DigitalSignature" ]
    path_length_constraint:
      description:
        - The number of levels allowed in a certificates path.
        - If this value is 0, the certificate does not have a restriction.
      returned: basic constraints extension exists
      type: int
      sample: 0
    public_key:
      description: The base64 encoded public key of the certificate.
      type: str
    cert_data:
      description: The base64 encoded data of the entire certificate.
      type: str
    serial_number:
      description: The serial number of the certificate represented as a hexadecimal string
      type: str
      sample: 01DEBCC4396DA010
    signature_algorithm:
      description: The algorithm used to create the certificate's signature
      type: str
      sample: sha1RSA
    ski:
      description: The certificate's subject key identifier
      returned: subject key identifier extension exists.
      type: str
      sample: 88271709A9B618608BECEBBAF64759C55254A3B7
    subject:
      description: The certificate's distinguished name.
      type: str
      sample: 'CN=Apple Worldwide Developer Relations Certification Authority, OU=Apple Worldwide Developer Relations, O=Apple Inc., C=US'
    thumbprint:
      description:
        - The thumbprint as a hex string of the certificate.
        - The return format will always be upper case.
      type: str
      sample: FF6797793A3CD798DC5B2ABEF56F73EDC9F83A64
    valid_from:
      description: The start date of the certificate represented in seconds since epoch.
      type: float
      sample: 1360255727
    valid_from_iso8601:
      description: The start date of the certificate represented as an iso8601 formatted date.
      type: str
      sample: '2017-12-15T08:39:32Z'
    valid_to:
      description: The expiry date of the certificate represented in seconds since epoch.
      type: float
      sample: 1675788527
    valid_to_iso8601:
      description: The expiry date of the certificate represented as an iso8601 formatted date.
      type: str
      sample: '2086-01-02T08:39:32Z'
    version:
      description: The x509 format version of the certificate
      type: int
      sample: 3
'''



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_ml_data_info
version_added: "3.16.0"
short_description: List ML Data resources
description:
    - List ML Data resources.
options:
    name:
        description:
            - Name of the data asset. If provided, all the data versions
              under this name will be returned.
        type: str
        required: false
    resource_group:
        description:
            - Name of resource group.
        required: false
        type: str
    ml_workspace:
        description:
            - Name of the Azure ML workspace.
        required: false
        type: str
    ml_registry:
        description:
            - If provided, the command will target the registry
              instead of a workspace.  Hence resource group and
              workspace won't be required. Must be provided if
              ml_workspace and resource_group are not provided.
        required: false
        type: str
    version:
        description:
            - Version of the data asset. Must be provided, if label
              is not provided. Mutually exclusive with label.
        required: false
        type: str
    label:
        description:
            - Label of the data asset. Must be provided, if version
              is not provided. Mutually exclusive with version.
        required: false
        type: str
    list_type:
        description:
            - Specify if you want active, archived or all datasets.
        type: str
        choices:
            - active
            - archived
            - all

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Bill Peck (@p3ck)
'''

EXAMPLES = '''
- name: List ML Data
  azure.azcollection.azure_rm_ml_data_info:
    name: MyData
    resource_group: myResourceGroup
    ml_workspace: myMLWorkspace

- name: List ML Data
  azure.azcollection.azure_rm_ml_data_info:
    resource_group: myResourceGroup
    ml_workspace: myMLWorkspace
    list_type: archived
'''

RETURN = '''
ml_data_assets:
    description:
        - Data that was just created or updated.
    returned: always
    type: dict
    sample: [
      {
      "creation_context":
        {
          "created_at": "2026-03-05T20:28:42.120932+00:00",
          "created_by": "ca413070-9a06-4e0a-8bb8-c167d29fa09e",
          "created_by_type": "Application",
          "last_modified_at": "2026-03-05T20:28:57.237566+00:00",
          "last_modified_by": "ca413070-9a06-4e0a-8bb8-c167d29fa09e",
          "last_modified_by_type": "Application",
        },
      "description": "Data asset created from local folder.",
      "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-data/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxx-xxx/data/data-xxxxxxx-xxx/versions/2",
      "name": "data-xxxxxxx-xxx",
      "path": "azureml://subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourcegroups/xxxxx-ml-data/workspaces/workspace-xxxxxxx-xxx/datastores/workspaceblobstore/paths/LocalUpload/xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx/data/",
      "properties": {},
      "tags": {},
      "type": "uri_folder",
      "version": "2",
      },
    ]
'''  # NOQA


try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ml import MLClientCommon
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMMLDataInfo(MLClientCommon):
    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(
                type='str',
                required=False,
            ),
            resource_group=dict(
                type='str',
                required=False,
            ),
            ml_workspace=dict(
                type='str',
                required=False,
            ),
            ml_registry=dict(
                type='str',
            ),
            version=dict(
                type='str',
            ),
            label=dict(
                type='str',
            ),
            list_type=dict(
                type='str',
                choices=['archived', 'active', 'all'],
            ),
        )

        self._client = None
        self.ml_data = None

        self.results = dict(
            ml_data_assets=[]
        )

        required_one_of = [('ml_workspace', 'ml_registry')]
        mutually_exclusive = [('version', 'label'),
                              ('ml_workspace', 'ml_registry'),
                              ('name', 'list_type')]
        required_together = [('ml_workspace', 'resource_group')]

        super(AzureRMMLDataInfo, self).__init__(self.module_arg_spec,
                                                supports_tags=False,
                                                supports_check_mode=True,
                                                required_one_of=required_one_of,
                                                mutually_exclusive=mutually_exclusive,
                                                required_together=required_together,
                                                facts_module=True,
                                                )

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        if self.name:
            label = self.label
            if not self.version and not label:
                label = 'latest'
            try:
                result = self.client.data.get(name=self.name,
                                              version=self.version,
                                              label=label)
                ml_data_assets = [self.entity_to_dict(result)]
            except ResourceNotFoundError:
                ml_data_assets = []
        else:
            list_view_type = self.get_list_view_type(self.list_type)
            results = self.client.data.list(list_view_type=list_view_type)

            ml_data_assets = [self.entity_to_dict(self.client.data.get(name=x.name, version=x.latest_version)) for x in results]

        self.results['ml_data_assets'] = ml_data_assets
        return self.results


def main():
    AzureRMMLDataInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_saml_roles
short_description: Resource module for organizations _saml _roles
description:
  - Manage operations create, update and delete of the resource organizations _saml
    _roles.
  - Create a SAML role.
  - Remove a SAML role.
  - Update a SAML role.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  networks:
    description: The list of networks that the SAML administrator has privileges on.
    elements: dict
    suboptions:
      access:
        description: The privilege of the SAML administrator on the network. Can be
          one of 'full', 'read-only', 'guest-ambassador', 'monitor-only', 'ssid-admin'
          or 'port-tags' or a custom role in the format custom-role ID.
        type: str
      id:
        description: The network ID.
        type: str
    type: list
  orgAccess:
    description: The privilege of the SAML administrator on the organization. Can
      be one of 'none', 'read-only', 'full' or 'enterprise' or a custom role in the
      format custom-role ID NAME.
    type: str
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
  role:
    description: The role of the SAML administrator.
    type: str
  samlRoleId:
    description: SamlRoleId path parameter. Saml role ID.
    type: str
  tags:
    description: The list of tags that the SAML administrator has privileges on.
    elements: dict
    suboptions:
      access:
        description: The privilege of the SAML administrator on the tag. Can be one
          of 'full', 'read-only', 'guest-ambassador' or 'monitor-only' or a custom
          role in the format custom-role ID.
        type: str
      tag:
        description: The name of the tag.
        type: str
    type: list
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations createOrganizationSamlRole
    description: Complete reference of the createOrganizationSamlRole API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-organization-saml-role
  - name: Cisco Meraki documentation for organizations deleteOrganizationSamlRole
    description: Complete reference of the deleteOrganizationSamlRole API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-organization-saml-role
  - name: Cisco Meraki documentation for organizations updateOrganizationSamlRole
    description: Complete reference of the updateOrganizationSamlRole API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-organization-saml-role
notes:
  - SDK Method used are
    organizations.Organizations.create_organization_saml_role,
    organizations.Organizations.delete_organization_saml_role,
    organizations.Organizations.update_organization_saml_role,
  - Paths used are
    post /organizations/{organizationId}/samlRoles,
    delete /organizations/{organizationId}/samlRoles/{samlRoleId},
    put /organizations/{organizationId}/samlRoles/{samlRoleId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_saml_roles:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    networks:
      - access: full
        id: N_24329156
    orgAccess: none
    organizationId: string
    role: myrole
    tags:
      - access: read-only
        tag: west
- name: Update by id
  cisco.meraki.organizations_saml_roles:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    networks:
      - access: full
        id: N_24329156
    orgAccess: none
    organizationId: string
    role: myrole
    samlRoleId: string
    tags:
      - access: read-only
        tag: west
- name: Delete by id
  cisco.meraki.organizations_saml_roles:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    organizationId: string
    samlRoleId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "id": "string",
      "role": "string",
      "orgAccess": "string",
      "networks": [
        {
          "id": "string",
          "access": "string"
        }
      ],
      "tags": [
        {
          "tag": "string",
          "access": "string"
        }
      ],
      "camera": [
        {
          "orgWide": true,
          "access": "string"
        }
      ]
    }
"""



#!/usr/bin/python
#
# Copyright (c) 2018 Zim Kalinowski, <zikalino@microsoft.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_containerregistry_info
version_added: "0.1.2"
short_description: Get Azure Container Registry facts
description:
    - Get facts for Container Registry.

options:
    resource_group:
        description:
            - The name of the resource group to which the container registry belongs.
        type: str
    name:
        description:
            - The name of the container registry.
        type: str
    retrieve_credentials:
        description:
            - Retrieve credentials for container registry.
        type: bool
        default: no
    tags:
        description:
            - List of tags to be matched.
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Zim Kalinowski (@zikalino)

'''

EXAMPLES = '''
- name: Get instance of Registry
  azure_rm_containerregistry_info:
    resource_group: myResourceGroup
    name: myRegistry

- name: List instances of Registry
  azure_rm_containerregistry_info:
    resource_group: myResourceGroup
    tags:
      - key
      - key:value
'''

RETURN = '''
registries:
    description:
        - A list of dictionaries containing facts for registries.
    returned: always
    type: complex
    contains:
        id:
            description:
                - The resource ID.
            returned: always
            type: str
            sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/Microsoft.ContainerRegistry/registr
                    ies/myRegistry"
        name:
            description:
                - The name of the resource.
            returned: always
            type: str
            sample: myRegistry
        location:
            description:
                - The location of the resource. This cannot be changed after the resource is created.
            returned: always
            type: str
            sample: westus
        admin_user_enabled:
            description:
                - Is admin user enabled.
            returned: always
            type: bool
            sample: yes
        sku:
            description:
                - The SKU name of the container registry.
            returned: always
            type: str
            sample: Premium
        provisioning_state:
            description:
                - Provisioning state of the container registry.
            returned: always
            type: str
            sample: Succeeded
        login_server:
            description:
                - Login server for the registry.
            returned: always
            type: str
            sample: acrd08521b.azurecr.io
        credentials:
            description:
                - Credentials, fields will be empty if admin user is not enabled for ACR.
            returned: when C(retrieve_credentials) is set and C(admin_user_enabled) is set on ACR
            type: complex
            contains:
                username:
                    description:
                        - The user name for container registry.
                    returned: when registry exists and C(admin_user_enabled) is set
                    type: str
                    sample: zim
                password:
                    description:
                        - password value.
                    returned: when registry exists and C(admin_user_enabled) is set
                    type: str
                    sample: pass1value
                password2:
                    description:
                        - password2 value.
                    returned: when registry exists and C(admin_user_enabled) is set
                    type: str
                    sample: pass2value
        identity:
            description:
                - Identity for the Server.
            type: complex
            returned: when available
            contains:
                type:
                    description:
                        - Type of the managed identity
                    returned: always
                    sample: UserAssigned
                    type: str
                user_assigned_identities:
                    description:
                        - User Assigned Managed Identities and its options
                    returned: always
                    type: complex
                    contains:
                        id:
                            description:
                                - Dict of the user assigned identities IDs associated to the Resource
                            returned: always
                            type: dict
                            elements: dict
        tags:
            description:
                - Tags assigned to the resource. Dictionary of string:string pairs.
            type: dict
            sample: { "tag1": "abc" }
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


class AzureRMContainerRegistryInfo(AzureRMModuleBase):
    def __init__(self):
        # define user inputs into argument
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str',
            ),
            name=dict(
                type='str'
            ),
            retrieve_credentials=dict(
                type='bool',
                default=False
            ),
            tags=dict(
                type='list',
                elements='str'
            )
        )
        # store the results of the module operation
        self.results = dict(
            changed=False
        )
        self.resource_group = None
        self.name = None
        self.retrieve_credentials = False
        self.tags = None

        super(AzureRMContainerRegistryInfo, self).__init__(self.module_arg_spec, supports_check_mode=True, supports_tags=False, facts_module=True)

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec) + ['tags']:
            setattr(self, key, kwargs[key])

        if self.name:
            self.results['registries'] = self.get()
        elif self.resource_group:
            self.results['registries'] = self.list_by_resource_group()
        else:
            self.results['registries'] = self.list_all()

        return self.results

    def get(self):
        response = None
        results = []
        try:
            response = self.containerregistry_client.registries.get(resource_group_name=self.resource_group,
                                                                    registry_name=self.name)
            self.log("Response : {0}".format(response))
        except Exception as e:
            self.log("Could not get facts for Registries: {0}".format(str(e)))

        if response is not None:
            if self.has_tags(response.tags, self.tags):
                results.append(self.format_item(response))

        return results

    def list_all(self):
        response = None
        results = []
        try:
            response = self.containerregistry_client.registries.list()
            self.log("Response : {0}".format(response))
        except Exception as e:
            self.fail("Could not get facts for Registries: {0}".format(str(e)))

        if response is not None:
            for item in response:
                if self.has_tags(item.tags, self.tags):
                    results.append(self.format_item(item))
        return results

    def list_by_resource_group(self):
        response = None
        results = []
        try:
            response = self.containerregistry_client.registries.list_by_resource_group(resource_group_name=self.resource_group)
            self.log("Response : {0}".format(response))
        except Exception as e:
            self.fail("Could not get facts for Registries: {0}".format(str(e)))

        if response is not None:
            for item in response:
                if self.has_tags(item.tags, self.tags):
                    results.append(self.format_item(item))
        return results

    def format_item(self, item):
        d = item.as_dict()
        resource_group = d['id'].split('resourceGroups/')[1].split('/')[0]
        name = d['name']
        credentials = {}
        admin_user_enabled = d['admin_user_enabled']

        if self.retrieve_credentials and admin_user_enabled:
            credentials = self.containerregistry_client.registries.list_credentials(resource_group_name=resource_group, registry_name=name).as_dict()
            for index in range(len(credentials['passwords'])):
                password = credentials['passwords'][index]
                if password['name'] == 'password':
                    credentials['password'] = password['value']
                elif password['name'] == 'password2':
                    credentials['password2'] = password['value']
            credentials.pop('passwords')

        d = {
            'resource_group': resource_group,
            'name': d['name'],
            'location': d['location'],
            'admin_user_enabled': admin_user_enabled,
            'sku': d['sku']['tier'].lower(),
            'provisioning_state': d['provisioning_state'],
            'login_server': d['login_server'],
            'id': d['id'],
            'tags': d.get('tags', None),
            'credentials': credentials,
            'identity': d.get('identity', None)
        }
        return d


def main():
    AzureRMContainerRegistryInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Micah Hunsberger (@mhunsber)
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_file_compression
short_description: Alters the compression of files and directories on NTFS partitions.
description:
  - This module sets the compressed attribute for files and directories on a filesystem that supports it like NTFS.
  - NTFS compression can be used to save disk space.
version_added: 2.7.0
options:
  path:
    description:
      - The full path of the file or directory to modify.
      - The path must exist on file system that supports compression like NTFS.
    required: yes
    type: path
  state:
    description:
      - Set to C(present) to ensure the I(path) is compressed.
      - Set to C(absent) to ensure the I(path) is not compressed.
    type: str
    choices:
      - absent
      - present
    default: present
  recurse:
    description:
      - Whether to recursively apply changes to all subdirectories and files.
      - This option only has an effect when I(path) is a directory.
      - When set to C(false), only applies changes to I(path).
      - When set to C(true), applies changes to I(path) and all subdirectories and files.
    type: bool
    default: false
  force:
    description:
      - This option only has an effect when I(recurse) is C(true)
      - If C(true), will check the compressed state of all subdirectories and files
        and make a change if any are different from I(compressed).
      - If C(false), will only make a change if the compressed state of I(path) is different from I(compressed).
      - If the folder structure is complex or contains a lot of files, it is recommended to set this
        option to C(false) so that not every file has to be checked.
    type: bool
    default: true
author:
  - Micah Hunsberger (@mhunsber)
notes:
  - M(ansible.windows.win_file_compression) sets the file system's compression state, it does not create a zip
    archive file.
  - For more about NTFS Compression, see U(http://www.ntfs.com/ntfs-compressed.htm)
'''

EXAMPLES = r'''
- name: Compress log files directory
  ansible.windows.win_file_compression:
    path: C:\Logs
    state: present

- name: Decompress log files directory
  ansible.windows.win_file_compression:
    path: C:\Logs
    state: absent

- name: Compress reports directory and all subdirectories
  ansible.windows.win_file_compression:
    path: C:\business\reports
    state: present
    recurse: true

# This will only check C:\business\reports for the compressed state
# If C:\business\reports is compressed, it will not make a change
# even if one of the child items is uncompressed

- name: Compress reports directory and all subdirectories (quick)
  ansible.windows.win_file_compression:
    path: C:\business\reports
    compressed: true
    recurse: true
    force: false
'''

RETURN = r'''
rc:
  description:
    - The return code of the compress/uncompress operation.
    - If no changes are made or the operation is successful, rc is 0.
  returned: always
  sample: 0
  type: int
'''



# Copyright (c) 2022, Gregory Furlong <gnfzdz@fzdz.io>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import os
import re
import typing as t

from ansible.module_utils.common.text.converters import to_bytes

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


def normalize_subvolume_path(path):
    """
    Normalizes btrfs subvolume paths to ensure exactly one leading slash, no trailing slashes and no consecutive slashes.
    In addition, if the path is prefixed with a leading <FS_TREE>, this value is removed.
    """
    fstree_stripped = re.sub(r"^<FS_TREE>", "", path)
    result = re.sub(r"/+$", "", re.sub(r"/+", "/", f"/{fstree_stripped}"))
    return result if len(result) > 0 else "/"


class BtrfsModuleException(Exception):
    pass


class BtrfsCommands:
    """
    Provides access to a subset of the Btrfs command line
    """

    def __init__(self, module: AnsibleModule) -> None:
        self.__module = module
        self.__module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}
        self.__btrfs: str = self.__module.get_bin_path("btrfs", required=True)

    def filesystem_show(self) -> list[dict[str, t.Any]]:
        command = [self.__btrfs, "filesystem", "show", "-d"]
        result = self.__module.run_command(command, check_rc=True)
        stdout = [x.strip() for x in result[1].splitlines()]
        filesystems = []
        current = None
        for line in stdout:
            if line.startswith("Label"):
                current = self.__parse_filesystem(line)
                filesystems.append(current)
            elif line.startswith("devid"):
                if current is None:
                    raise ValueError("Found 'devid' line without previous 'Label' line")
                current["devices"].append(self.__parse_filesystem_device(line))
        return filesystems

    def __parse_filesystem(self, line) -> dict[str, t.Any]:
        label = re.sub(r"\s*uuid:.*$", "", re.sub(r"^Label:\s*", "", line))
        id = re.sub(r"^.*uuid:\s*", "", line)

        filesystem: dict[str, t.Any] = {}
        filesystem["label"] = label.strip("'") if label != "none" else None
        filesystem["uuid"] = id
        filesystem["devices"] = []
        filesystem["mountpoints"] = []
        filesystem["subvolumes"] = []
        filesystem["default_subvolid"] = None
        return filesystem

    def __parse_filesystem_device(self, line: str) -> str:
        return re.sub(r"^.*path\s", "", line)

    def subvolumes_list(self, filesystem_path: str) -> list[dict[str, t.Any]]:
        command = [self.__btrfs, "subvolume", "list", "-tap", to_bytes(filesystem_path)]
        result = self.__module.run_command(command, check_rc=True)
        stdout = [x.split("\t") for x in result[1].splitlines()]
        subvolumes: list[dict[str, t.Any]] = [{"id": 5, "parent": None, "path": "/"}]
        if len(stdout) > 2:
            subvolumes.extend([self.__parse_subvolume_list_record(x) for x in stdout[2:]])
        return subvolumes

    def __parse_subvolume_list_record(self, item: list[str]) -> dict[str, t.Any]:
        return {
            "id": int(item[0]),
            "parent": int(item[2]),
            "path": normalize_subvolume_path(item[5]),
        }

    def subvolume_get_default(self, filesystem_path: str) -> int:
        command = [self.__btrfs, "subvolume", "get-default", to_bytes(filesystem_path)]
        result = self.__module.run_command(command, check_rc=True)
        # ID [n] ...
        return int(result[1].strip().split()[1])

    def subvolume_set_default(self, filesystem_path: str, subvolume_id: int) -> None:
        command = [self.__btrfs, "subvolume", "set-default", str(subvolume_id), to_bytes(filesystem_path)]
        self.__module.run_command(command, check_rc=True)

    def subvolume_create(self, subvolume_path: str) -> None:
        command = [self.__btrfs, "subvolume", "create", to_bytes(subvolume_path)]
        self.__module.run_command(command, check_rc=True)

    def subvolume_snapshot(self, snapshot_source: str, snapshot_destination: str) -> None:
        command = [self.__btrfs, "subvolume", "snapshot", to_bytes(snapshot_source), to_bytes(snapshot_destination)]
        self.__module.run_command(command, check_rc=True)

    def subvolume_delete(self, subvolume_path: str) -> None:
        command = [self.__btrfs, "subvolume", "delete", to_bytes(subvolume_path)]
        self.__module.run_command(command, check_rc=True)


class BtrfsInfoProvider:
    """
    Utility providing details of the currently available btrfs filesystems
    """

    def __init__(self, module: AnsibleModule) -> None:
        self.__module = module
        self.__btrfs_api = BtrfsCommands(module)
        self.__findmnt_path: str = self.__module.get_bin_path("findmnt", required=True)

    def get_filesystems(self) -> list[dict[str, t.Any]]:
        filesystems = self.__btrfs_api.filesystem_show()
        mountpoints = self.__find_mountpoints()
        for filesystem in filesystems:
            device_mountpoints = self.__filter_mountpoints_for_devices(mountpoints, filesystem["devices"])
            filesystem["mountpoints"] = device_mountpoints

            if len(device_mountpoints) > 0:
                # any path within the filesystem can be used to query metadata
                mountpoint = device_mountpoints[0]["mountpoint"]
                filesystem["subvolumes"] = self.get_subvolumes(mountpoint)
                filesystem["default_subvolid"] = self.get_default_subvolume_id(mountpoint)

        return filesystems

    def get_mountpoints(self, filesystem_devices: list[str]) -> list[dict[str, t.Any]]:
        mountpoints = self.__find_mountpoints()
        return self.__filter_mountpoints_for_devices(mountpoints, filesystem_devices)

    def get_subvolumes(self, filesystem_path) -> list[dict[str, t.Any]]:
        return self.__btrfs_api.subvolumes_list(filesystem_path)

    def get_default_subvolume_id(self, filesystem_path) -> int:
        return self.__btrfs_api.subvolume_get_default(filesystem_path)

    def __filter_mountpoints_for_devices(
        self, mountpoints: list[dict[str, t.Any]], devices: list[str]
    ) -> list[dict[str, t.Any]]:
        return [m for m in mountpoints if (m["device"] in devices)]

    def __find_mountpoints(self) -> list[dict[str, t.Any]]:
        command = [self.__findmnt_path, "-t", "btrfs", "-nvP"]
        result = self.__module.run_command(command)
        mountpoints = []
        if result[0] == 0:
            lines = result[1].splitlines()
            for line in lines:
                mountpoint = self.__parse_mountpoint_pairs(line)
                mountpoints.append(mountpoint)
        return mountpoints

    def __parse_mountpoint_pairs(self, line) -> dict[str, t.Any]:
        pattern = re.compile(
            r'^TARGET="(?P<target>.*)"\s+SOURCE="(?P<source>.*)"\s+FSTYPE="(?P<fstype>.*)"\s+OPTIONS="(?P<options>.*)"\s*$'
        )
        match = pattern.search(line)
        if match is not None:
            groups = match.groupdict()

            return {
                "mountpoint": groups["target"],
                "device": groups["source"],
                "subvolid": self.__extract_mount_subvolid(groups["options"]),
            }
        else:
            raise BtrfsModuleException(f"Failed to parse findmnt result for line: {line!r}")

    def __extract_mount_subvolid(self, mount_options: str) -> int:
        for option in mount_options.split(","):
            if option.startswith("subvolid="):
                return int(option[len("subvolid=") :])
        raise BtrfsModuleException(f"Failed to find subvolid for mountpoint in options {mount_options!r}")


class BtrfsSubvolume:
    """
    Wrapper class providing convenience methods for inspection of a btrfs subvolume
    """

    def __init__(self, filesystem: BtrfsFilesystem, subvolume_id: int):
        self.__filesystem = filesystem
        self.__subvolume_id = subvolume_id

    def get_filesystem(self) -> BtrfsFilesystem:
        return self.__filesystem

    def is_mounted(self) -> bool:
        mountpoints = self.get_mountpoints()
        return mountpoints is not None and len(mountpoints) > 0

    def is_filesystem_root(self) -> bool:
        return self.__subvolume_id == 5

    def is_filesystem_default(self) -> bool:
        return self.__filesystem.default_subvolid == self.__subvolume_id

    def get_mounted_path(self) -> str | None:
        mountpoints = self.get_mountpoints()
        if mountpoints is not None and len(mountpoints) > 0:
            return mountpoints[0]
        if self.parent is not None:
            parent = self.__filesystem.get_subvolume_by_id(self.parent)
            parent_path = parent.get_mounted_path() if parent else None
            if parent_path is not None:
                return f"{parent_path}{os.path.sep}{self.name}"
        return None

    def get_mountpoints(self) -> list[str]:
        return self.__filesystem.get_mountpoints_by_subvolume_id(self.__subvolume_id)

    def get_child_relative_path(self, absolute_child_path: str) -> str:
        """
        Get the relative path from this subvolume to the named child subvolume.
        The provided parameter is expected to be normalized as by normalize_subvolume_path.
        """
        path = self.path
        if absolute_child_path.startswith(path):
            relative = absolute_child_path[len(path) :]
            return re.sub(r"^/*", "", relative)
        else:
            raise BtrfsModuleException(f"Path '{absolute_child_path}' doesn't start with '{path}'")

    def get_parent_subvolume(self) -> BtrfsSubvolume | None:
        parent_id = self.parent
        return self.__filesystem.get_subvolume_by_id(parent_id) if parent_id is not None else None

    def get_child_subvolumes(self) -> list[BtrfsSubvolume]:
        return self.__filesystem.get_subvolume_children(self.__subvolume_id)

    @property
    def __info(self) -> dict[str, t.Any]:
        result = self.__filesystem.get_subvolume_info_for_id(self.__subvolume_id)
        # assert result is not None
        return result  # type: ignore

    @property
    def id(self) -> int:
        return self.__subvolume_id

    @property
    def name(self):
        return self.path.split("/").pop()

    @property
    def path(self) -> str:
        return self.__info["path"]

    @property
    def parent(self):
        return self.__info["parent"]


class BtrfsFilesystem:
    """
    Wrapper class providing convenience methods for inspection of a btrfs filesystem
    """

    def __init__(self, info: dict[str, t.Any], provider: BtrfsInfoProvider, module: AnsibleModule) -> None:
        self.__provider = provider

        # constant for module execution
        self.__uuid: str = info["uuid"]
        self.__label: str = info["label"]
        self.__devices: list[str] = info["devices"]

        # refreshable
        self.__default_subvolid: int | None = info["default_subvolid"] if "default_subvolid" in info else None
        self.__update_mountpoints(info["mountpoints"] if "mountpoints" in info else [])
        self.__update_subvolumes(info["subvolumes"] if "subvolumes" in info else [])

    @property
    def uuid(self) -> str:
        return self.__uuid

    @property
    def label(self) -> str:
        return self.__label

    @property
    def default_subvolid(self) -> int | None:
        return self.__default_subvolid

    @property
    def devices(self) -> list[str]:
        return list(self.__devices)

    def refresh(self) -> None:
        self.refresh_mountpoints()
        self.refresh_subvolumes()
        self.refresh_default_subvolume()

    def refresh_mountpoints(self) -> None:
        mountpoints = self.__provider.get_mountpoints(list(self.__devices))
        self.__update_mountpoints(mountpoints)

    def __update_mountpoints(self, mountpoints: list[dict[str, t.Any]]) -> None:
        self.__mountpoints: dict[int, list[str]] = dict()
        for i in mountpoints:
            subvolid: int = i["subvolid"]
            mountpoint: str = i["mountpoint"]
            if subvolid not in self.__mountpoints:
                self.__mountpoints[subvolid] = []
            self.__mountpoints[subvolid].append(mountpoint)

    def refresh_subvolumes(self) -> None:
        filesystem_path = self.get_any_mountpoint()
        if filesystem_path is not None:
            subvolumes = self.__provider.get_subvolumes(filesystem_path)
            self.__update_subvolumes(subvolumes)

    def __update_subvolumes(self, subvolumes: list[dict[str, t.Any]]) -> None:
        # TODO strategy for retaining information on deleted subvolumes?
        self.__subvolumes: dict[int, dict[str, t.Any]] = dict()
        for subvolume in subvolumes:
            self.__subvolumes[subvolume["id"]] = subvolume

    def refresh_default_subvolume(self) -> None:
        filesystem_path = self.get_any_mountpoint()
        if filesystem_path is not None:
            self.__default_subvolid = self.__provider.get_default_subvolume_id(filesystem_path)

    def contains_device(self, device: str) -> bool:
        return device in self.__devices

    def contains_subvolume(self, subvolume: str) -> bool:
        return self.get_subvolume_by_name(subvolume) is not None

    def get_subvolume_by_id(self, subvolume_id: int) -> BtrfsSubvolume | None:
        return BtrfsSubvolume(self, subvolume_id) if subvolume_id in self.__subvolumes else None

    def get_subvolume_info_for_id(self, subvolume_id: int) -> dict[str, t.Any] | None:
        return self.__subvolumes[subvolume_id] if subvolume_id in self.__subvolumes else None

    def get_subvolume_by_name(self, subvolume: str) -> BtrfsSubvolume | None:
        for subvolume_info in self.__subvolumes.values():
            if subvolume_info["path"] == subvolume:
                return BtrfsSubvolume(self, subvolume_info["id"])
        return None

    def get_any_mountpoint(self) -> str | None:
        for subvol_mountpoints in self.__mountpoints.values():
            if len(subvol_mountpoints) > 0:
                return subvol_mountpoints[0]
        # maybe error?
        return None

    def get_any_mounted_subvolume(self) -> BtrfsSubvolume | None:
        for subvolid, subvol_mountpoints in self.__mountpoints.items():
            if len(subvol_mountpoints) > 0:
                return self.get_subvolume_by_id(subvolid)
        return None

    def get_mountpoints_by_subvolume_id(self, subvolume_id: int) -> list[str]:
        return self.__mountpoints[subvolume_id] if subvolume_id in self.__mountpoints else []

    def get_nearest_subvolume(self, subvolume: str) -> BtrfsSubvolume:
        """Return the identified subvolume if existing, else the closest matching parent"""
        subvolumes_by_path = self.__get_subvolumes_by_path()
        while len(subvolume) > 1:
            if subvolume in subvolumes_by_path:
                return BtrfsSubvolume(self, subvolumes_by_path[subvolume]["id"])
            else:
                subvolume = re.sub(r"/[^/]+$", "", subvolume)

        return BtrfsSubvolume(self, 5)

    def get_mountpath_as_child(self, subvolume_name: str) -> str:
        """Find a path to the target subvolume through a mounted ancestor"""
        nearest = self.get_nearest_subvolume(subvolume_name)
        nearest_or_none: BtrfsSubvolume | None = nearest
        if nearest.path == subvolume_name:
            nearest_or_none = nearest.get_parent_subvolume()
        if nearest_or_none is None or nearest_or_none.get_mounted_path() is None:
            raise BtrfsModuleException(f"Failed to find a path '{subvolume_name}' through a mounted parent subvolume")
        else:
            return f"{nearest_or_none.get_mounted_path()}{os.path.sep}{nearest_or_none.get_child_relative_path(subvolume_name)}"

    def get_subvolume_children(self, subvolume_id: int) -> list[BtrfsSubvolume]:
        return [BtrfsSubvolume(self, x["id"]) for x in self.__subvolumes.values() if x["parent"] == subvolume_id]

    def __get_subvolumes_by_path(self) -> dict[str, dict[str, t.Any]]:
        result = {}
        for s in self.__subvolumes.values():
            path = s["path"]
            result[path] = s
        return result

    def is_mounted(self) -> bool:
        return self.__mountpoints is not None and len(self.__mountpoints) > 0

    def get_summary(self) -> dict[str, t.Any]:
        subvolumes = []
        sources = self.__subvolumes.values() if self.__subvolumes is not None else []
        for subvolume in sources:
            id = subvolume["id"]
            subvolumes.append(
                {
                    "id": id,
                    "path": subvolume["path"],
                    "parent": subvolume["parent"],
                    "mountpoints": self.get_mountpoints_by_subvolume_id(id),
                }
            )

        return {
            "default_subvolume": self.__default_subvolid,
            "devices": self.__devices,
            "label": self.__label,
            "uuid": self.__uuid,
            "subvolumes": subvolumes,
        }


class BtrfsFilesystemsProvider:
    """
    Provides methods to query available btrfs filesystems
    """

    def __init__(self, module: AnsibleModule) -> None:
        self.__module = module
        self.__provider = BtrfsInfoProvider(module)
        self.__filesystems: dict[str, BtrfsFilesystem] | None = None

    def get_matching_filesystem(self, criteria: dict[str, t.Any]) -> BtrfsFilesystem:
        if criteria["device"] is not None:
            criteria["device"] = os.path.realpath(criteria["device"])

        self.__check_init()
        # assert self.__filesystems is not None # TODO
        self_filesystems: dict[str, BtrfsFilesystem] = self.__filesystems  # type: ignore
        matching = [f for f in self_filesystems.values() if self.__filesystem_matches_criteria(f, criteria)]
        if len(matching) == 1:
            return matching[0]
        else:
            raise BtrfsModuleException(
                f"Found {len(matching)} filesystems matching criteria uuid={criteria['uuid']} label={criteria['label']} device={criteria['device']}"
            )

    def __filesystem_matches_criteria(self, filesystem: BtrfsFilesystem, criteria: dict[str, t.Any]):
        return (
            (criteria["uuid"] is None or filesystem.uuid == criteria["uuid"])
            and (criteria["label"] is None or filesystem.label == criteria["label"])
            and (criteria["device"] is None or filesystem.contains_device(criteria["device"]))
        )

    def get_filesystem_for_device(self, device: str) -> BtrfsFilesystem | None:
        real_device = os.path.realpath(device)
        self.__check_init()
        # assert self.__filesystems is not None # TODO
        self_filesystems: dict[str, BtrfsFilesystem] = self.__filesystems  # type: ignore
        for fs in self_filesystems.values():
            if fs.contains_device(real_device):
                return fs
        return None

    def get_filesystems(self) -> list[BtrfsFilesystem]:
        self.__check_init()
        # assert self.__filesystems is not None # TODO
        self_filesystems: dict[str, BtrfsFilesystem] = self.__filesystems  # type: ignore
        return list(self_filesystems.values())

    def __check_init(self) -> None:
        if self.__filesystems is None:
            self.__filesystems = dict()
            for f in self.__provider.get_filesystems():
                uuid = f["uuid"]
                self.__filesystems[uuid] = BtrfsFilesystem(f, self.__provider, self.__module)



#!/usr/bin/python

# Copyright (C) 2018 IBM CORPORATION
# Author(s): Tzur Eliyahu <tzure@il.ibm.com>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ibm_sa_pool
short_description: Handles pools on IBM Spectrum Accelerate Family storage systems

description:
  - This module creates or deletes pools to be used on IBM Spectrum Accelerate Family storage systems.
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  pool:
    description:
      - Pool name.
    required: true
    type: str
  state:
    description:
      - Pool state.
    default: "present"
    choices: ["present", "absent"]
    type: str
  size:
    description:
      - Pool size in GB.
    type: str
  snapshot_size:
    description:
      - Pool snapshot size in GB.
    type: str
  domain:
    description:
      - Adds the pool to the specified domain.
    type: str
  perf_class:
    description:
      - Assigns a perf_class to the pool.
    type: str

extends_documentation_fragment:
  - community.general._ibm_storage
  - community.general._attributes

author:
  - Tzur Eliyahu (@tzure)
"""

EXAMPLES = r"""
- name: Create new pool.
  community.general.ibm_sa_pool:
    name: pool_name
    size: 300
    state: present
    username: admin
    password: secret
    endpoints: hostdev-system

- name: Delete pool.
  community.general.ibm_sa_pool:
    name: pool_name
    state: absent
    username: admin
    password: secret
    endpoints: hostdev-system
"""
RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._ibm_sa_utils import (
    connect_ssl,
    execute_pyxcli_command,
    is_pyxcli_installed,
    spectrum_accelerate_spec,
)


def main():
    argument_spec = spectrum_accelerate_spec()
    argument_spec.update(
        dict(
            state=dict(default="present", choices=["present", "absent"]),
            pool=dict(required=True),
            size=dict(),
            snapshot_size=dict(),
            domain=dict(),
            perf_class=dict(),
        )
    )

    module = AnsibleModule(argument_spec)

    is_pyxcli_installed(module)

    xcli_client = connect_ssl(module)
    pool = xcli_client.cmd.pool_list(pool=module.params["pool"]).as_single_element
    state = module.params["state"]

    state_changed = False
    if state == "present" and not pool:
        state_changed = execute_pyxcli_command(module, "pool_create", xcli_client)
    if state == "absent" and pool:
        state_changed = execute_pyxcli_command(module, "pool_delete", xcli_client)

    module.exit_json(changed=state_changed)


if __name__ == "__main__":
    main()



#
# (c) 2016 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function


__metaclass__ = type

from ansible.module_utils.connection import Connection
from ansible.utils.display import Display
from ansible_collections.ansible.netcommon.plugins.action.network import (
    ActionModule as ActionNetworkModule,
)


display = Display()


class ActionModule(ActionNetworkModule):
    def run(self, tmp=None, task_vars=None):
        del tmp  # tmp no longer has any effect

        module_name = self._task.action.split(".")[-1]
        self._config_module = True if module_name in ["nxos_config", "config"] else False
        persistent_connection = self._play_context.connection.split(".")[-1]

        if (self._play_context.connection == "httpapi") and module_name in (
            "nxos_file_copy",
            "nxos_nxapi",
        ):
            return {
                "failed": True,
                "msg": f"Connection httpapi is not valid for {module_name} module.",
            }

        if module_name == "nxos_file_copy":
            # when file_pull is enabled, the file_pull_timeout and connect_ssh_port options
            # will override persistent_command_timeout and port
            # this has been kept for backwards compatibility till these options are removed
            if persistent_connection != "network_cli":
                return {
                    "failed": True,
                    "msg": (
                        f"Connection type must be fully qualified name for "
                        f"network_cli connection type, got {self._play_context.connection}"
                    ),
                }

            conn = Connection(self._connection.socket_path)
            persistent_command_timeout = conn.get_option("persistent_command_timeout")
            file_pull = self._task.args.get("file_pull", False)
            file_pull_timeout = self._task.args.get("file_pull_timeout")
            connect_ssh_port = self._task.args.get("connect_ssh_port", 22)

            if file_pull:
                # if file_pull_timeout is explicitly set, use that
                if file_pull_timeout:
                    conn.set_option("persistent_command_timeout", file_pull_timeout)
                # if file_pull_timeout is not set and command_timeout < 300s, bump to 300s.
                elif persistent_command_timeout < 300:
                    conn.set_option("persistent_command_timeout", 300)
                conn.set_option("port", connect_ssh_port)

        if module_name == "nxos_install_os":
            connection = self._connection
            persistent_command_timeout = connection.get_option(
                "persistent_command_timeout",
            )
            persistent_connect_timeout = connection.get_option(
                "persistent_connect_timeout",
            )

            display.vvvv(
                f"PERSISTENT_COMMAND_TIMEOUT is {persistent_command_timeout}",
                self._play_context.remote_addr,
            )
            display.vvvv(
                f"PERSISTENT_CONNECT_TIMEOUT is %s {persistent_connect_timeout}",
                self._play_context.remote_addr,
            )
            if persistent_command_timeout < 600 or persistent_connect_timeout < 600:
                msg = "PERSISTENT_COMMAND_TIMEOUT and PERSISTENT_CONNECT_TIMEOUT"
                msg += " must be set to 600 seconds or higher when using nxos_install_os module."
                msg += " Current persistent_command_timeout setting:" + str(
                    persistent_command_timeout,
                )
                msg += " Current persistent_connect_timeout setting:" + str(
                    persistent_connect_timeout,
                )
                return {"failed": True, "msg": msg}

        if persistent_connection in ("network_cli", "httpapi"):
            if module_name == "nxos_gir":
                conn = Connection(self._connection.socket_path)
                persistent_command_timeout = conn.get_option(
                    "persistent_command_timeout",
                )
                gir_timeout = 200
                if persistent_command_timeout < gir_timeout:
                    conn.set_option("persistent_command_timeout", gir_timeout)
                    msg = f"timeout value extended to %ss for nxos_gir {gir_timeout}"
                    display.warning(msg)

        else:
            return {
                "failed": True,
                "msg": f"Connection type {self._play_context.connection} is not valid for this module",
            }

        return super(ActionModule, self).run(task_vars=task_vars)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_vpc_egress_igw
version_added: 1.0.0
version_added_collection: community.aws
short_description: Manage an AWS VPC Egress Only Internet gateway
description:
  - Manage an AWS VPC Egress Only Internet gateway
author:
  - Daniel Shepherd (@shepdelacreme)
options:
  vpc_id:
    description:
      - The VPC ID for the VPC that this Egress Only Internet Gateway should be attached.
    required: true
    type: str
  state:
    description:
      - Create or delete the EIGW.
    default: present
    choices: [ 'present', 'absent' ]
    type: str
notes:
  - Support for O(tags) and O(purge_tags) was added in release 9.0.0.
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
  - amazon.aws.tags.modules
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

# Ensure that the VPC has an Internet Gateway.
# The Internet Gateway ID is can be accessed via {{eigw.gateway_id}} for use in setting up NATs etc.
- name: Create Egress internet only gateway
  amazon.aws.ec2_vpc_egress_igw:
    vpc_id: vpc-abcdefgh
    state: present

- name: Delete Egress internet only gateway
  amazon.aws.ec2_vpc_egress_igw:
    vpc_id: vpc-abcdefgh
    state: absent
"""

RETURN = r"""
gateway_id:
    description: The ID of the Egress Only Internet Gateway or Null.
    returned: always
    type: str
    sample: "eigw-0e00cf111ba5bc11e"
vpc_id:
    description: The ID of the VPC to attach or detach gateway from.
    returned: always
    type: str
    sample: "vpc-012345678"
tags:
    description: Any tags assigned to the internet gateway.
    returned: always
    type: dict
"""

from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AnsibleEC2Error
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import create_egress_only_internet_gateway
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import delete_egress_only_internet_gateway
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_egress_only_internet_gateways
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import ensure_ec2_tags
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict


def delete_eigw(module: AnsibleAWSModule, connection, eigw_id: str) -> Dict[str, Union[str, bool]]:
    """
    Delete EIGW.

    module     : AnsibleAWSModule object
    connection : boto3 client connection object
    eigw_id    : ID of the EIGW to delete
    """

    if module.check_mode:
        return dict(
            changed=True, msg=f"Would have deleted Egress internet only Gateway id '{eigw_id}' if not in check mode."
        )

    try:
        changed = delete_egress_only_internet_gateway(connection, egress_only_internet_gateway_id=eigw_id)
    except AnsibleEC2Error as e:
        module.fail_json_aws(e)

    return dict(changed=changed)


def create_eigw(module: AnsibleAWSModule, connection, vpc_id: str) -> Dict[str, Union[str, bool]]:
    """
    Create EIGW.

    module       : AnsibleAWSModule object
    connection   : boto3 client connection object
    vpc_id       : ID of the VPC we are operating on
    """

    if module.check_mode:
        return dict(changed=True, msg="Would have created Egress internet only Gateway if not in check mode.")

    gateway_id = None
    changed = False

    try:
        response = create_egress_only_internet_gateway(connection, vpc_id=vpc_id, tags=module.params.get("tags"))
        changed = True
    except AnsibleEC2Error as e:
        module.fail_json_aws(e)

    gateway = response.get("EgressOnlyInternetGateway", {})
    state = gateway.get("Attachments", [{}])[0].get("State")
    gateway_id = gateway.get("EgressOnlyInternetGatewayId")
    tags = boto3_tag_list_to_ansible_dict(gateway.get("Tags", []))

    if not gateway_id or state not in ("attached", "attaching"):
        # EIGW gave back a bad attachment state or an invalid response so we error out
        module.fail_json(
            msg=f"Unable to create and attach Egress Only Internet Gateway to VPCId: {vpc_id}. Bad or no state in response",
            **camel_dict_to_snake_dict(response),
        )

    return dict(changed=changed, gateway_id=gateway_id, tags=tags)


def find_egress_only_igw(module: AnsibleAWSModule, connection, vpc_id: str) -> Optional[Dict[str, Any]]:
    """
    Describe EIGWs.

    module     : AnsibleAWSModule object
    connection : boto3 client connection object
    vpc_id     : ID of the VPC we are operating on
    """
    result = None

    try:
        for eigw in describe_egress_only_internet_gateways(connection):
            for attachment in eigw.get("Attachments", []):
                if attachment.get("VpcId") == vpc_id and attachment.get("State") in ("attached", "attaching"):
                    return {
                        "gateway_id": eigw.get("EgressOnlyInternetGatewayId"),
                        "tags": boto3_tag_list_to_ansible_dict(eigw.get("Tags", [])),
                    }
    except AnsibleEC2Error as e:
        module.fail_json_aws(e)

    return result


def ensure_present(connection, module: AnsibleAWSModule, existing: Optional[Dict[str, Any]]) -> None:
    vpc_id = module.params.get("vpc_id")
    result = dict(vpc_id=vpc_id, changed=False)

    if not existing:
        result.update(create_eigw(module, connection, vpc_id))
    else:
        egress_only_igw_id = existing.get("gateway_id")
        changed = False
        result = existing
        tags = module.params.get("tags")
        purge_tags = module.params.get("purge_tags")
        if tags is not None:
            changed = ensure_ec2_tags(
                connection,
                module,
                egress_only_igw_id,
                resource_type="egress-only-internet-gateway",
                tags=tags,
                purge_tags=purge_tags,
            )
        result.update(dict(changed=changed, vpc_id=vpc_id))

    module.exit_json(**result)


def ensure_absent(connection, module: AnsibleAWSModule, existing: Optional[Dict[str, Any]]) -> None:
    vpc_id = module.params.get("vpc_id")
    if not existing:
        module.exit_json(changed=False, msg=f"No Egress only internet gateway attached to the VPC id '{vpc_id}'")

    egress_only_igw_id = existing.get("gateway_id")
    result = dict(gateway_id=egress_only_igw_id, vpc_id=vpc_id, changed=False)
    result.update(delete_eigw(module, connection, egress_only_igw_id))
    module.exit_json(**result)


def main():
    argument_spec = dict(
        vpc_id=dict(required=True),
        state=dict(default="present", choices=["present", "absent"]),
        tags=dict(type="dict", aliases=["resource_tags"]),
        purge_tags=dict(type="bool", default=True),
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True)

    connection = module.client("ec2")

    vpc_id = module.params.get("vpc_id")
    state = module.params.get("state")

    existing_egress_only_igw = find_egress_only_igw(module, connection, vpc_id)

    if state == "present":
        ensure_present(connection, module, existing_egress_only_igw)
    else:
        ensure_absent(connection, module, existing_egress_only_igw)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2024 xuzhang3 (@xuzhang3), Fred-sun (@Fred-sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_postgresqlflexibledatabase_info
version_added: "2.2.0"
short_description: Get Azure PostgreSQL Flexible Database facts
description:
    - Get facts of PostgreSQL Flexible Database.

options:
    resource_group:
        description:
            - The name of the resource group that contains the resource. You can obtain this value from the Azure Resource Manager API or the portal.
        type: str
        required: True
    server_name:
        description:
            - The name of the post gresql server.
        type: str
        required: True
    name:
        description:
            - The name of the post gresql database.
        type: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)

'''

EXAMPLES = '''
- name: List instance of PostgreSQL Flexible Database by server name
  azure_rm_postgresqlflexibledatabase_info:
    resource_group: myResourceGroup
    server_name: server_name

- name: Get instances of PostgreSQL Flexible Database
  azure_rm_postgresqlflexibledatabase_info:
    resource_group: myResourceGroup
    server_name: server_name
    name: database_name
'''

RETURN = '''
database:
    description:
        - A list of dictionaries containing facts for PostgreSQL Flexible Database.
    returned: always
    type: complex
    contains:
        id:
            description:
                - Resource ID of the postgresql flexible database.
            returned: always
            type: str
            sample: "/subscriptions/xxx-xxx/resourceGroups/testRG/providers/Microsoft.DBforPostgreSQL/flexibleServers/postfle9/databases/freddatabase"
        name:
            description:
                - Resource name.
            returned: always
            type: str
            sample: freddatabase
        charset:
            description:
                - The charset of the database.
            returned: always
            type: str
            sample: UTF-8
        collation:
            description:
                - The collation of the database.
            returned: always
            type: str
            sample: en_US.utf8
        type:
            description:
                - The type of the resource.
            returned: always
            type: str
            sample: Microsoft.DBforPostgreSQL/flexibleServers/databases
        system_data:
            description:
                - The system metadata relating to this resource.
            type: complex
            returned: always
            contains:
                created_by:
                    description:
                        - The identity that created the resource.
                    type: str
                    returned: always
                    sample: null
                created_by_type:
                    description:
                        - The type of identity that created the resource.
                    returned: always
                    type: str
                    sample: null
                created_at:
                    description:
                        - The timestamp of resource creation (UTC).
                    returned: always
                    sample: null
                    type: str
                last_modified_by:
                    description:
                        - The identity that last modified the resource.
                    type: str
                    returned: always
                    sample: null
                last_modified_by_type:
                    description:
                        - The type of identity that last modified the resource.
                    returned: always
                    sample: null
                    type: str
                last_modified_at:
                    description:
                        - The timestamp of resource last modification (UTC).
                    returned: always
                    sample: null
                    type: str
'''


try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMPostgreSqlFlexibleDatabaseInfo(AzureRMModuleBase):
    def __init__(self):
        # define user inputs into argument
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str',
                required=True
            ),
            server_name=dict(
                type='str',
                required=True
            ),
            name=dict(
                type='str'
            ),
        )
        # store the results of the module operation
        self.results = dict(
            changed=False
        )
        self.resource_group = None
        self.name = None
        self.server_name = None
        super(AzureRMPostgreSqlFlexibleDatabaseInfo, self).__init__(self.module_arg_spec, supports_check_mode=True, supports_tags=False, facts_module=True)

    def exec_module(self, **kwargs):
        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        if self.name is not None:
            self.results['databases'] = self.get()
        else:
            self.results['databases'] = self.list_all()
        return self.results

    def get(self):
        response = None
        results = []
        try:
            response = self.postgresql_flexible_client.databases.get(resource_group_name=self.resource_group,
                                                                     server_name=self.server_name,
                                                                     database_name=self.name)
            self.log("Response : {0}".format(response))
        except ResourceNotFoundError:
            self.log('Could not get facts for PostgreSQL Flexible Server.')

        if response is not None:
            results.append(self.format_item(response))

        return results

    def list_all(self):
        response = None
        results = []
        try:
            response = self.postgresql_flexible_client.databases.list_by_server(resource_group_name=self.resource_group,
                                                                                server_name=self.server_name)
            self.log("Response : {0}".format(response))
        except Exception as ec:
            self.log('Could not get facts for PostgreSQL Flexible Servers.')

        if response is not None:
            for item in response:
                results.append(self.format_item(item))

        return results

    def format_item(self, item):
        result = dict(
            id=item.id,
            name=item.name,
            system_data=dict(),
            type=item.type,
            charset=item.charset,
            collation=item.collation
        )
        if item.system_data is not None:
            result['system_data']['created_by'] = item.system_data.created_by
            result['system_data']['created_by_type'] = item.system_data.created_by_type
            result['system_data']['created_at'] = item.system_data.created_at
            result['system_data']['last_modified_by'] = item.system_data.last_modified_by
            result['system_data']['last_modified_by_type'] = item.system_data.last_modified_by_type
            result['system_data']['last_modified_at'] = item.system_data.last_modified_at

        return result


def main():
    AzureRMPostgreSqlFlexibleDatabaseInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_spot_instance
version_added: 2.0.0
short_description: Request, stop, reboot or cancel spot instance
description:
    - Creates or cancels spot instance requests.
author:
  - Sri Rachana Achyuthuni (@srirachanaachyuthuni)
options:
  zone_group:
    description:
      - Name for logical grouping of spot requests.
      - All spot instances in the request are launched in the same availability zone.
    type: str
  client_token:
    description: The idempotency token you provided when you launched the instance, if applicable.
    type: str
  count:
    description:
      - Number of instances to launch.
    default: 1
    type: int
  interruption:
    description:
      - The behavior when a Spot Instance is interrupted.
    choices: [ "hibernate", "stop", "terminate" ]
    type: str
    default: terminate
  launch_group:
    description:
      - Launch group for spot requests, see U(https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/how-spot-instances-work.html#spot-launch-group).
    type: str
  launch_specification:
    description:
      - The launch specification.
    type: dict
    suboptions:
      security_group_ids:
        description:
          - Security group id (or list of ids) to use with the instance.
        type: list
        elements: str
      security_groups:
        description:
          - Security group name (or list of group names) to use with the instance.
          - Only supported with EC2 Classic. To launch in a VPC, use O(launch_specification.security_group_ids).
        type: list
        elements: str
      key_name:
        description:
          - Key to use on the instance.
          - The SSH key must already exist in AWS in order to use this argument.
          - Keys can be created / deleted using the M(amazon.aws.ec2_key) module.
        type: str
      subnet_id:
        description:
          - The ID of the subnet in which to launch the instance.
        type: str
      user_data:
        description:
          - The base64-encoded user data for the instance. User data is limited to 16 KB.
        type: str
      block_device_mappings:
        description:
          - A list of hash/dictionaries of volumes to add to the new instance.
        type: list
        elements: dict
        suboptions:
          device_name:
            description:
              - The device name (for example, /dev/sdh or xvdh).
            type: str
          virtual_name:
            description:
              - The virtual device name
            type: str
          ebs:
            description:
              - Parameters used to automatically set up EBS volumes when the instance is launched,
                see U(https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.request_spot_instances).
            type: dict
          no_device:
            description:
              - To omit the device from the block device mapping, specify an empty string.
            type: str
      ebs_optimized:
        description:
          - Whether instance is using optimized EBS volumes, see U(https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/EBSOptimized.html).
        default: false
        type: bool
      iam_instance_profile:
        description:
          - The IAM instance profile.
        type: dict
        suboptions:
          arn:
            description:
              - The Amazon Resource Name (ARN) of the instance profile.
              - Only one of O(launch_specification.iam_instance_profile.arn) or I(launch_specification.iam_instance_profile.name) may be specified.
            type: str
          name:
            description:
              - The name of the instance profile.
              - Only one of O(launch_specification.iam_instance_profile.arn) or I(launch_specification.iam_instance_profile.name) may be specified.
            type: str
      image_id:
        description:
          - The ID of the AMI.
        type: str
      instance_type:
        description:
          - Instance type to use for the instance, see U(https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instance-types.html).
          - Required when creating a new instance.
        type: str
      kernel_id:
        description:
          - The ID of the kernel.
        type: str
      network_interfaces:
        description:
          - One or more network interfaces. If you specify a network interface, you must specify subnet IDs and security group IDs using the network interface.
        type: list
        elements: dict
        default: []
        suboptions:
          associate_public_ip_address:
            description:
              - Indicates whether to assign a public IPv4 address to an instance you launch in a VPC.
            type: bool
          delete_on_termination:
            description:
              - If set to true , the interface is deleted when the instance is terminated.
                You can specify true only if creating a new network interface when launching an instance.
            type: bool
          description:
            description:
              - The description of the network interface. Applies only if creating a network interface when launching an instance.
            type: str
          device_index:
            description:
              - The position of the network interface in the attachment order. A primary network interface has a device index of 0.
              - If you specify a network interface when launching an instance, you must specify the device index.
            type: int
          groups:
            description:
              - The IDs of the security groups for the network interface. Applies only if creating a network interface when launching an instance.
            type: list
            elements: str
          ipv6_address_count:
            description:
              - A number of IPv6 addresses to assign to the network interface
            type: int
          ipv6_addresses:
            description:
              - One or more IPv6 addresses to assign to the network interface.
            type: list
            elements: dict
            suboptions:
              ipv6address:
                description: The IPv6 address.
                type: str
          network_interface_id:
            description:
              - The ID of the network interface.
            type: str
          private_ip_address:
            description:
              - The private IPv4 address of the network interface
            type: str
          private_ip_addresses:
            description:
              - One or more private IPv4 addresses to assign to the network interface
            type: list
            elements: dict
          secondary_private_ip_address_count:
            description:
              - The number of secondary private IPv4 addresses.
            type: int
          subnet_id:
            description:
              - The ID of the subnet associated with the network interface
            type: str
          associate_carrier_ip_address:
            description:
              - Indicates whether to assign a carrier IP address to the network interface.
            type: bool
          interface_type:
            description:
              - The type of network interface.
            type: str
            choices: ['interface', 'efa']
          network_card_index:
            description:
              - The index of the network card.
            type: int
          ipv4_prefixes:
            description:
              - One or more IPv4 delegated prefixes to be assigned to the network interface.
            type: list
            elements: dict
          ipv4_prefix_count:
            description:
              - The number of IPv4 delegated prefixes to be automatically assigned to the network interface
            type: int
          ipv6_prefixes:
            description:
              - One or more IPv6 delegated prefixes to be assigned to the network interface
            type: list
            elements: dict
          ipv6_prefix_count:
            description:
              - The number of IPv6 delegated prefixes to be automatically assigned to the network interface
            type: int
      placement:
        description:
          - The placement information for the instance.
        type: dict
        suboptions:
          availability_zone:
            description:
              - The Availability Zone.
            type: str
          group_name:
            description:
              - The name of the placement group.
            type: str
          tenancy:
            description:
              - the tenancy of the host
            type: str
            choices: ['default', 'dedicated', 'host']
            default: default
      ramdisk_id:
        description:
          - The ID of the RAM disk.
        type: str
      monitoring:
        description:
          - Indicates whether basic or detailed monitoring is enabled for the instance.
        type: dict
        suboptions:
          enabled:
            description:
              - Indicates whether detailed monitoring is enabled. Otherwise, basic monitoring is enabled.
            type: bool
            default: false
  state:
    description:
      - Whether the spot request should be created or removed.
      - When O(state=present), O(launch_specification) is required.
      - When O(state=absent), O(spot_instance_request_ids) is required.
    default: 'present'
    choices: [ 'absent', 'present' ]
    type: str
  spot_price:
    description:
      - Maximum spot price to bid. If not set, a regular on-demand instance is requested.
      - A spot request is made with this maximum bid. When it is filled, the instance is started.
    type: str
  spot_type:
    description:
      - The type of spot request.
      - After being interrupted a V(persistent) spot instance will be started once there is capacity to fill the request again.
    default: 'one-time'
    choices: [ "one-time", "persistent" ]
    type: str
  tags:
    description:
      - A dictionary of key-value pairs for tagging the Spot Instance request on creation.
    type: dict
  spot_instance_request_ids:
    description:
      - List of strings with IDs of spot requests to be cancelled
    type: list
    elements: str
  terminate_instances:
    description:
      - Boolean value to set whether or not to terminate instances associated to spot request.
      - Can be used only when O(state=absent).
    default: False
    type: bool
    version_added: 5.4.0
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Simple Spot Request Creation
  amazon.aws.ec2_spot_instance:
    launch_specification:
      image_id: ami-123456789
      key_name: my-keypair
      instance_type: t2.medium

- name: Spot Request Creation with more options
  amazon.aws.ec2_spot_instance:
    launch_specification:
      image_id: ami-123456789
      key_name: my-keypair
      instance_type: t2.medium
      subnet_id: subnet-12345678
      block_device_mappings:
        - device_name: /dev/sdb
          ebs:
            delete_on_termination: true
            volume_type: gp3
            volume_size: 5
        - device_name: /dev/sdc
          ebs:
            delete_on_termination: true
            volume_type: io2
            volume_size: 30
      network_interfaces:
        - associate_public_ip_address: false
          delete_on_termination: true
          device_index: 0
      placement:
        availability_zone: us-west-2a
      monitoring:
        enabled: false
    spot_price: 0.002
    tags:
      Environment: Testing

- name: Spot Request Termination
  amazon.aws.ec2_spot_instance:
    spot_instance_request_ids: ['sir-12345678', 'sir-abcdefgh']
    state: absent
"""

RETURN = r"""
spot_request:
    description: The spot instance request details after creation.
    returned: when success
    type: dict
    sample: {
        "create_time": "2021-08-23T22:59:12+00:00",
        "instance_interruption_behavior": "terminate",
        "launch_specification": {
            "block_device_mappings": [
                {
                    "device_name": "/dev/sdb",
                    "ebs": {
                        "delete_on_termination": true,
                        "volume_size": 5,
                        "volume_type": "gp3"
                    }
                }
            ],
            "ebs_optimized": false,
            "iam_instance_profile": {
                "arn": "arn:aws:iam::EXAMPLE:instance-profile/myinstanceprofile"
            },
            "image_id": "ami-083ac7c7ecf9bb9b0",
            "instance_type": "t2.small",
            "key_name": "mykey",
            "monitoring": {
                "enabled": false
            },
            "network_interfaces": [
                {
                    "associate_public_ip_address": false,
                    "delete_on_termination": true,
                    "device_index": 0
                }
            ],
            "placement": {
                "availability_zone": "us-west-2a",
                "tenancy": "default"
            },
            "security_groups": [
                {
                    "group_name": "default"
                }
            ]
        },
        "product_description": "Linux/UNIX",
        "spot_instance_request_id": "sir-1234abcd",
        "spot_price": "0.00600",
        "state": "open",
        "status": {
            "code": "pending-evaluation",
            "message": "Your Spot request has been submitted for review, and is pending evaluation.",
            "update_time": "2021-08-23T22:59:12+00:00"
        },
        "tags": {},
        "type": "one-time"
        }

cancelled_spot_request:
    description: Details of the cancelled spot instance request.
    type: str
    returned: when cancellation is successful
    sample: 'Spot requests with IDs: sir-1234abcd have been cancelled'
"""

# TODO: add support for datetime-based parameters
# import datetime
# import time

from typing import Any
from typing import Dict

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict
from ansible.module_utils.common.dict_transformations import snake_dict_to_camel_dict

from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AnsibleEC2Error
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import cancel_spot_instance_requests
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_spot_instance_requests
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import request_spot_instances
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import terminate_instances
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict


def build_launch_specification(launch_spec: Dict[str, Any]) -> Dict[str, Any]:
    """
    Remove keys that have a value of None from Launch Specification
    Descend into these subkeys:
    network_interfaces
    block_device_mappings
    monitoring
    placement
    iam_instance_profile
    """
    assigned_keys = dict((k, v) for k, v in launch_spec.items() if v is not None)

    sub_key_to_build = ["placement", "iam_instance_profile", "monitoring"]
    for subkey in sub_key_to_build:
        if launch_spec[subkey] is not None:
            assigned_keys[subkey] = dict((k, v) for k, v in launch_spec[subkey].items() if v is not None)

    if launch_spec["network_interfaces"] is not None:
        interfaces = []
        for iface in launch_spec["network_interfaces"]:
            interfaces.append(dict((k, v) for k, v in iface.items() if v is not None))
        assigned_keys["network_interfaces"] = interfaces

    if launch_spec["block_device_mappings"] is not None:
        block_devs = []
        for dev in launch_spec["block_device_mappings"]:
            block_devs.append(dict((k, v) for k, v in dev.items() if v is not None))
        assigned_keys["block_device_mappings"] = block_devs

    return snake_dict_to_camel_dict(assigned_keys, capitalize_first=True)


def _request(module: AnsibleAWSModule, connection) -> None:
    # request_spot_instances() always creates a new spot request
    changed = True

    if module.check_mode:
        module.exit_json(changed=changed)

    params = {}

    if module.params.get("launch_specification"):
        params["LaunchSpecification"] = build_launch_specification(module.params.get("launch_specification"))

    if module.params.get("zone_group"):
        params["AvailabilityZoneGroup"] = module.params.get("zone_group")

    if module.params.get("count"):
        params["InstanceCount"] = module.params.get("count")

    if module.params.get("launch_group"):
        params["LaunchGroup"] = module.params.get("launch_group")

    if module.params.get("spot_price"):
        params["SpotPrice"] = module.params.get("spot_price")

    if module.params.get("spot_type"):
        params["Type"] = module.params.get("spot_type")

    if module.params.get("client_token"):
        params["ClientToken"] = module.params.get("client_token")

    if module.params.get("interruption"):
        params["InstanceInterruptionBehavior"] = module.params.get("interruption")

    if module.params.get("tags"):
        params["TagSpecifications"] = [
            {
                "ResourceType": "spot-instances-request",
                "Tags": ansible_dict_to_boto3_tag_list(module.params.get("tags")),
            }
        ]

    # TODO: add support for datetime-based parameters
    # params['ValidFrom'] = module.params.get('valid_from')
    # params['ValidUntil'] = module.params.get('valid_until')

    try:
        request_spot_instance_response = request_spot_instances(connection, **params)[0]
    except AnsibleEC2Error as e:
        module.fail_json_aws(e, msg="Error while creating the spot instance request")

    request_spot_instance_response["Tags"] = boto3_tag_list_to_ansible_dict(
        request_spot_instance_response.get("Tags", [])
    )
    spot_request = camel_dict_to_snake_dict(request_spot_instance_response, ignore_list=["Tags"])
    module.exit_json(spot_request=spot_request, changed=changed)


def _cancel(module: AnsibleAWSModule, connection) -> None:
    changed = False
    spot_instance_request_ids = module.params.get("spot_instance_request_ids")
    try:
        requests_exist = describe_spot_instance_requests(
            connection,
            SpotInstanceRequestIds=spot_instance_request_ids,
            Filters=[{"Name": "state", "Values": ["open", "active"]}],
        )
    except AnsibleEC2Error as e:
        module.fail_json_aws(e, msg="Failure when describing spot requests")

    if len(requests_exist) > 0:
        changed = True
        if module.check_mode:
            module.exit_json(changed=changed, msg=f"Would have cancelled Spot request {spot_instance_request_ids}")

        try:
            cancel_spot_instance_requests(connection, spot_instance_request_ids)
        except AnsibleEC2Error as e:
            module.fail_json_aws(e, msg="Error while cancelling the spot instance request")

        if module.params.get("terminate_instances") is True:
            instance_ids = [request["InstanceId"] for request in requests_exist]
            try:
                terminate_instances(connection, instance_ids=instance_ids)
            except AnsibleEC2Error as e:
                module.fail_json_aws(e, msg=f"Failed to terminate instances with Ids={instance_ids}")

        module.exit_json(changed=changed, msg=f"Cancelled Spot request {spot_instance_request_ids}")
    else:
        module.exit_json(changed=changed, msg="Spot request not found or already cancelled")


def main() -> None:
    network_interface_options = dict(
        associate_public_ip_address=dict(type="bool"),
        delete_on_termination=dict(type="bool"),
        description=dict(type="str"),
        device_index=dict(type="int"),
        groups=dict(type="list", elements="str"),
        ipv6_address_count=dict(type="int"),
        ipv6_addresses=dict(type="list", elements="dict", options=dict(ipv6address=dict(type="str"))),
        network_interface_id=dict(type="str"),
        private_ip_address=dict(type="str"),
        private_ip_addresses=dict(type="list", elements="dict"),
        secondary_private_ip_address_count=dict(type="int"),
        subnet_id=dict(type="str"),
        associate_carrier_ip_address=dict(type="bool"),
        interface_type=dict(type="str", choices=["interface", "efa"]),
        network_card_index=dict(type="int"),
        ipv4_prefixes=dict(type="list", elements="dict"),
        ipv4_prefix_count=dict(type="int"),
        ipv6_prefixes=dict(type="list", elements="dict"),
        ipv6_prefix_count=dict(type="int"),
    )
    block_device_mappings_options = dict(
        device_name=dict(type="str"),
        virtual_name=dict(type="str"),
        ebs=dict(type="dict"),
        no_device=dict(type="str"),
    )
    monitoring_options = dict(enabled=dict(type="bool", default=False))
    placement_options = dict(
        availability_zone=dict(type="str"),
        group_name=dict(type="str"),
        tenancy=dict(type="str", choices=["default", "dedicated", "host"], default="default"),
    )
    iam_instance_profile_options = dict(arn=dict(type="str"), name=dict(type="str"))
    launch_specification_options = dict(
        security_group_ids=dict(type="list", elements="str"),
        security_groups=dict(type="list", elements="str"),
        block_device_mappings=dict(type="list", elements="dict", options=block_device_mappings_options),
        ebs_optimized=dict(type="bool", default=False),
        iam_instance_profile=dict(type="dict", options=iam_instance_profile_options),
        image_id=dict(type="str"),
        instance_type=dict(type="str"),
        kernel_id=dict(type="str"),
        key_name=dict(type="str"),
        monitoring=dict(type="dict", options=monitoring_options),
        network_interfaces=dict(type="list", elements="dict", options=network_interface_options, default=[]),
        placement=dict(type="dict", options=placement_options),
        ramdisk_id=dict(type="str"),
        user_data=dict(type="str"),
        subnet_id=dict(type="str"),
    )

    argument_spec = dict(
        zone_group=dict(type="str"),
        client_token=dict(type="str", no_log=False),
        count=dict(type="int", default=1),
        interruption=dict(type="str", default="terminate", choices=["hibernate", "stop", "terminate"]),
        launch_group=dict(type="str"),
        launch_specification=dict(type="dict", options=launch_specification_options),
        state=dict(default="present", choices=["present", "absent"]),
        spot_price=dict(type="str"),
        spot_type=dict(default="one-time", choices=["one-time", "persistent"]),
        tags=dict(type="dict"),
        # valid_from=dict(type='datetime', default=datetime.datetime.now()),
        # valid_until=dict(type='datetime', default=(datetime.datetime.now() + datetime.timedelta(minutes=60))
        spot_instance_request_ids=dict(type="list", elements="str"),
        terminate_instances=dict(type="bool", default="False"),
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True)

    state = module.params["state"]

    if module.params.get("terminate_instances") and state != "absent":
        module.fail_json("terminate_instances can only be used when state is absent.")

    connection = module.client("ec2")

    if state == "present":
        _request(module, connection)

    if state == "absent":
        _cancel(module, connection)


if __name__ == "__main__":
    main()
